"""WHAD server tool

This utility implements a server module, allowing to create a TCP proxy
which can be used to access a device remotely.
"""
import os
import logging
import time
import traceback
import string
from argparse import ArgumentParser, Namespace

from whad.exceptions import RequiredImplementation, UnsupportedDomain, UnsupportedCapability
from whad.cli.ui import wait, warning, error, success, info, display_packet
from whad.cli.app import CommandLineApp, run_app
from whad.device import Bridge, ProtocolHub
from scapy.all import *
from whad.scapy.layers.rf4ce import *
from whad.scapy.layers.phy import *
from whad.scapy.layers.esb import *
from whad.scapy.layers.unifying import *
from whad.device.unix import UnixConnector, UnixSocketServerDevice
from whad.tools.utils import list_implemented_injectors, get_injector_parameters, gen_option_name, build_configuration_from_args
#from whad.unifying import Injector
from whad.phy.connector.injector import Injector

from queue import Queue
logger = logging.getLogger(__name__)



class WhadDomainSubParser(ArgumentParser):
    """
    Implements a Whad Domain subparser.
    """
    def warning(self, message):
        """Display a warning message in orange (if color is enabled)
        """
        warning(message)

    def error(self, message):
        """Display an error message in red (if color is enabled)
        """
        error(message)
        #exit(1)

class WhadInjectApp(CommandLineApp):
    connector = None
    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD generic injection tool',
            interface=True,
            commands=False,
            input=CommandLineApp.INPUT_WHAD,
            output=CommandLineApp.OUTPUT_WHAD
        )

        self.add_argument(
            '-r',
            '--repeat',
            dest='repeat',
            action='store_true',
            default=False,
            help="Repeat the transmission of packets"
        )


        self.add_argument(
            '-d',
            '--delay',
            dest='delay',
            default=0,
            help="Delay between the transmission of two consecutive packets"
        )

        self._input_queue = Queue()
        self._repeat_queue = Queue()

        # Don't build yet if piped
        if not self.is_stdin_piped():
            self.subparsers = self.add_subparsers(
                required=True,
                dest="domain",
                parser_class=WhadDomainSubParser,
                help='Domain in use'
            )

            self.build_subparsers(self.subparsers)
        else:
            self.subparsers = None

    def build_subparsers(self, subparsers):
        """
        Generate the subparsers argument according to the environment.
        """
        # List every domain implementing an injector
        self.environment = list_implemented_injectors()

        # Iterate over domain, and get the associated sniffer parameters
        for domain_name in self.environment:
            self.environment[domain_name]["parameters"] = get_injector_parameters(
                self.environment[domain_name]["configuration_class"]
            )


            self.environment[domain_name]["subparser"] = subparsers.add_parser(
                domain_name,
                description="WHAD {} Injection tool".format(domain_name.capitalize())
            )

            # Iterate over every parameters, and add arguments to subparsers
            for (
                    parameter_name,
                    (parameter_type, parameter_default, parameter_base_class, parameter_help)
                ) in self.environment[domain_name]["parameters"].items():

                dest = parameter_name

                # If parameter is based on a dataclass, process a subparameter
                if parameter_base_class is not None:
                    parameter_base, parameter_name = parameter_name.split(".")

                parameter_name = gen_option_name(parameter_name)

                # Process parameter help and shortname
                if parameter_help is not None and "(" in parameter_help:
                    parameter_shortnames = [
                        "-{}".format(i) for i in
                        parameter_help.split("(")[1].replace(")","").split(",")
                    ]
                    parameter_help = parameter_help.split("(")[0]
                else:
                    parameter_shortnames = []


                # Process parameter type
                if parameter_type != bool:
                    choices = []
                    # If we got an int
                    if parameter_type == int:
                        # allow to provide hex arguments
                        parameter_type = lambda x: int(x,0)
                    # If we got a list, it is a list of string
                    if parameter_type == list:
                        parameter_default=[]
                        parameter_type=str
                        action = "append"
                    elif parameter_type == bytes:
                        parameter_type = bytes.fromhex
                        action = "store"
                    else:
                        action = "store"
                    # Add non-boolean argument to corresponding subparser
                    self.environment[domain_name]["subparser"].add_argument(
                        "--"+parameter_name,
                        *parameter_shortnames,
                        default=parameter_default,
                        action=action,
                        type=parameter_type,
                        dest=dest,
                        help=parameter_help
                    )
                else:
                    # Add boolean argument to corresponding subparser
                    self.environment[domain_name]["subparser"].add_argument(
                        "--"+parameter_name,
                        *parameter_shortnames,
                        action='store_true',
                        dest=dest,
                        help=parameter_help
                    )

            self.environment[domain_name]["subparser"].add_argument(
                dest='packet',
                nargs="*",
                help='packet to inject'
            )

    def on_incoming_packet(self, packet):
        self._input_queue.put(packet)

    def generate_packets(self):
        self.provided_count = 0
        if hasattr(self.args, "packet"):
            for p in self.args.packet:
                try:
                    pkt = eval(p)
                    self._input_queue.put(pkt)
                    self.provided_count += 1
                except Exception as e:
                    # If hexadecimal only, try to interpret as bytes
                    if all(c in string.hexdigits for c in p):
                        try:
                            pkt = bytes.fromhex(p)
                            self._input_queue.put(pkt)
                            self.provided_count += 1
                        except Exception as e:
                            error("Failure during packet interpretation: " + str(p) + " -> " + str(e))
                    else:
                        error("Failure during packet interpretation: " + str(p) + " -> " + str(e))

    def pre_run(self):
        """Pre-run operations: rewrite arguments to allow skipping domain name when a pipe is configured.
        """

        try:
            try:
                index = sys.argv.index("--interface")
            except ValueError:
                index = sys.argv.index("-i")
        except ValueError:
            if "-h" in sys.argv or "--help" in sys.argv:
                self.print_help()
                exit(1)
            error("You need to provide an interface.")
            exit(1)

        start_argv = sys.argv[:index + 2]
        end_argv =  sys.argv[index + 2:]

        if len(sys.argv) == 1 or "-h" in start_argv or "--help" in start_argv:
            self.print_help()
            exit(1)


        error_func = self.error
        if self.subparsers is None:
            self.error  = lambda m : None
        super().pre_run()
        try:
            domain = self.args.domain
        except AttributeError:
            self.error("You have to provide a domain.")
            exit(1)

        if self.subparsers is None:
            self.error  = error_func

        if self.subparsers is None:
            sys.argv = start_argv

            self.subparsers = self.add_subparsers(
                required=True,
                dest="domain",
                parser_class=WhadDomainSubParser,
                help='Domain in use'
            )

            self.build_subparsers(self.subparsers)

            if domain not in start_argv and domain not in end_argv:
                sys.argv = start_argv + [domain] + end_argv
            else:
                sys.argv = start_argv + end_argv

            for k, v in self.parse_args().__dict__.items():
                setattr(self.args, k, v)

        # If no color is not selected, configure scapy color theme
        if not self.args.nocolor:
            conf.color_theme = BrightTheme()

        self.generate_packets()

    def run(self):
        monitors = []
        injector = None
        # Launch pre-run tasks
        self.pre_run()

        try:
            # We need to have an interface specified
            if self.interface is not None:
                # We need to have a domain specified
                if self.args.domain is not None:
                    # Parse the arguments to populate a injector configuration
                    configuration = build_configuration_from_args(self.environment, self.args)

                    # Generate an injector based on the selected domain
                    injector = self.environment[self.args.domain]["injector_class"](self.interface)


                    if self.is_piped_interface():

                        connector = UnixConnector(self.input_interface)

                        connector.domain = self.args.domain
                        hub = ProtocolHub(2)
                        connector.format = hub.get(self.args.domain).format
                        connector.on_packet = self.on_incoming_packet

                    try:
                        injector.configuration = configuration
                    except Exception as e:
                        self.error("Error during configuration: " +repr(e))

                    while True:
                        while (self.input_interface is not None and self.input_interface.opened) or self.provided_count > 0 or not self._input_queue.empty():
                            if not self._input_queue.empty():
                                packet = self._input_queue.get()
                                try:
                                    if isinstance(packet, bytes):
                                        packet = injector.format(packet)[0]
                                        # patch for BLE only
                                        if BTLE_RF in packet:
                                            packet = packet[BTLE]
                                    info("Transmitting: ")
                                    display_packet(packet)
                                    injector.inject(packet)
                                    if float(self.args.delay) > 0:
                                        time.sleep(float(self.args.delay))
                                    if self.args.repeat:
                                        self._repeat_queue.put(packet)
                                except Exception as e:
                                    self.error("Error during injection: " + repr(e))
                                    traceback.print_exc()
                                self.provided_count-=1

                        if not self.args.repeat:
                            break

                        while not self._repeat_queue.empty():
                            self._input_queue.put(self._repeat_queue.get())
                            self.provided_count += 1
                else:
                    self.error("You need to specify a domain.")
            else:
                self.error('You need to specify an interface with option --interface.')

        except UnsupportedDomain as unsupported_domain:
            self.error('WHAD device doesn\'t support selected domain ({})'.format(self.args.domain))

        except UnsupportedCapability as unsupported_capability:
            self.error('WHAD device doesn\'t support selected capability ({})'.format(unsupported_capability.capability))

        except KeyboardInterrupt as keybd:
            self.warning('injector stopped (CTRL-C)')
            injector.stop()
            injector.close()
            exit()


def winject_main():
    app = WhadInjectApp()
    run_app(app)



from whad.common.converters.hid.mappings import HID_SPECIALS, HID_MAP
from whad.common.converters.hid.exceptions import HIDKeyNotFound, HIDLocaleNotFound, HIDCodeNotFound

class HIDConverter:
    '''
    This class provides a basic API to convert an HID code to an human friendly keystroke and vice versa.
    '''
    @classmethod
    def get_hid_code_from_key(self, key, alt=False, ctrl=False, shift=False, gui=False, locale="fr"):
        '''
        This function converts a key to a tuple composed of HID code and HID modifiers.

        :param key: main key to convert
        :param alt: alt key is pressed
        :param ctrl: ctrl key is pressed
        :param shift: shift key is pressed
        :param gui: gui key is pressed

        '''
        if locale not in HID_MAP:
            raise HIDLocaleNotFound(locale)

        if key in HID_MAP[locale]:
            hid_code, modifiers = HID_MAP[locale][key]
        elif key in HID_SPECIALS:
            hid_code, modifiers = HID_SPECIALS[key]
        else:
            raise HIDKeyNotFound(key, alt, ctrl, shift, gui)

        special_keys = {
            "ALT": alt,
            "CTRL" : ctrl,
            "SHIFT" : shift,
            "GUI" : gui
        }

        for special_key, pressed in special_keys.items():
            if pressed:
                modifiers += HID_SPECIALS[special_key][1]

        return (hid_code, modifiers)

    @classmethod
    def get_key_from_hid_code(self, hid_code = 0, modifiers = 0, locale="fr"):
        '''
        This function converts an HID code and HID modifiers to the corresponding keystroke.

        :param hid_code: HID code to convert
        :param hid_modifiers: HID modifiers to convert
        '''

        for key, value in HID_MAP[locale].items():
            if value == (hid_code, modifiers):
                return key

        for key, value in HID_SPECIALS.items():
            if value == (hid_code, modifiers):
                return key

        raise HIDCodeNotFound(hid_code, modifiers)



import ctypes
import functools
import socket
import struct
import select
from ctypes import sizeof
from scapy.data import MTU
from scapy.consts import WINDOWS
from scapy.supersocket import SuperSocket
from scapy.packet import bind_layers, Packet
from scapy.fields import BitField
from scapy.error import warning
from scapy.layers.bluetooth import BluetoothUserSocket, BluetoothSocketError, BluetoothCommandError, \
    HCI_Hdr, SM_Hdr

class SM_Security_Request(Packet):
    name = "Security Request"
    fields_desc = [
       BitField("authentication", 0, 8)
    ]

bind_layers(SM_Hdr, SM_Security_Request, sm_command=0x0b)

class BluetoothUserSocketFixed(BluetoothUserSocket):
    pass

'''
class sockaddr_hci(ctypes.Structure):
    _fields_ = [
        ("sin_family", ctypes.c_ushort),
        ("hci_dev", ctypes.c_ushort),
        ("hci_channel", ctypes.c_ushort),
    ]

class BluetoothUserSocketFixed(SuperSocket):
    desc = "read/write H4 over a Bluetooth user channel"

    def __init__(self, adapter_index=0):
        if WINDOWS:
            warning("Not available on Windows")
            return
        # s = socket.socket(socket.AF_BLUETOOTH, socket.SOCK_RAW, socket.BTPROTO_HCI)  # noqa: E501
        # s.bind((0,1))

        # yeah, if only
        # thanks to Python's weak ass socket and bind implementations, we have
        # to call down into libc with ctypes

        sockaddr_hcip = ctypes.POINTER(sockaddr_hci)
        ctypes.cdll.LoadLibrary("libc.so.6")
        libc = ctypes.CDLL("libc.so.6")

        socket_c = libc.socket
        socket_c.argtypes = (ctypes.c_int, ctypes.c_int, ctypes.c_int)
        socket_c.restype = ctypes.c_int

        bind = libc.bind
        bind.argtypes = (ctypes.c_int,
                         ctypes.POINTER(sockaddr_hci),
                         ctypes.c_int)
        bind.restype = ctypes.c_int

        ########
        # actual code

        s = socket_c(31, 3, 1)  # (AF_BLUETOOTH, SOCK_RAW, HCI_CHANNEL_USER)
        if s < 0:
            raise BluetoothSocketError("Unable to open PF_BLUETOOTH socket")
        self.hci_fd = s

        sa = sockaddr_hci()
        sa.sin_family = 31  # AF_BLUETOOTH
        sa.hci_dev = adapter_index  # adapter index
        sa.hci_channel = 1   # HCI_USER_CHANNEL

        r = bind(s, sockaddr_hcip(sa), sizeof(sa))
        if r != 0:
            raise BluetoothSocketError("Unable to bind")

        self.ins = self.outs = socket.fromfd(s, 31, 3, 1)

    def send_command(self, cmd):
        opcode = cmd.opcode
        self.send(cmd)
        while True:
            r = self.recv()
            if r.type == 0x04 and r.code == 0xe and r.opcode == opcode:
                if r.status != 0:
                    raise BluetoothCommandError("Command %x failed with %x" % (opcode, r.status))  # noqa: E501
                return r

    def recv(self, x=MTU):
        return HCI_Hdr(self.ins.recv(x))

    def readable(self, timeout=0):
        (ins, outs, foo) = select.select([self.ins], [], [], timeout)
        return len(ins) > 0

    def flush(self):
        while self.readable():
            self.recv()

    def close(self):
        if self.closed:
            return

        # Properly close socket so we can free the device
        ctypes.cdll.LoadLibrary("libc.so.6")
        libc = ctypes.CDLL("libc.so.6")

        close = libc.close
        close.restype = ctypes.c_int
        self.closed = True
        if hasattr(self, "outs"):
            if not hasattr(self, "ins") or self.ins != self.outs:
                if self.outs and (WINDOWS or self.outs.fileno() != -1):
                    close(self.outs.fileno())
        if hasattr(self, "ins"):
            if self.ins and (WINDOWS or self.ins.fileno() != -1):
                close(self.ins.fileno())
        close(self.hci_fd)
'''



from whad.ble.connector import BLE, Injector, Hijacker
from whad.ble.utils.phy import is_access_address_valid
from whad.ble.sniffing import SynchronizedConnection, SnifferConfiguration, AccessAddress, \
    SynchronizationEvent, DesynchronizationEvent, KeyExtractedEvent
from whad.ble.crypto import EncryptedSessionInitialization, LinkLayerDecryptor, LegacyPairingCracking
from whad.hub.ble import AccessAddressDiscovered, Synchronized, BleRawPduReceived, BlePduReceived, BleAdvPduReceived
from whad.ble.exceptions import MissingCryptographicMaterial
from whad.exceptions import WhadDeviceDisconnected
from whad.ble import UnsupportedCapability, message_filter
from scapy.layers.bluetooth4LE import BTLE_DATA, BTLE
from whad.common.sniffing import EventsManager
from struct import pack
from time import sleep
import logging

logger = logging.getLogger(__name__)

class Sniffer(BLE, EventsManager):
    """
    BLE Sniffer interface for compatible WHAD device.
    """

    def __init__(self, device):
        BLE.__init__(self, device)
        EventsManager.__init__(self)
        self.__synchronized = False
        self.__connection = None
        self.__access_addresses = {}
        self.__decryptor = LinkLayerDecryptor()
        self.__encrypted_session_initialization = EncryptedSessionInitialization()
        self.__legacy_pairing_cracking = LegacyPairingCracking()
        self.__configuration = SnifferConfiguration()

        # Check if device accepts advertisements or connection sniffing
        if not self.can_sniff_advertisements() and not self.can_sniff_new_connection():
            raise UnsupportedCapability("Sniff")

    @property
    def synchronized(self):
        return self.__synchronized

    def wait_new_connection(self, address="FF:FF:FF:FF:FF:FF"):
        self.filter = address
        self.configure(advertisements=False, connection=True)
        self.start()
        while not self.synchronized:
            sleep(0.01)
        return self.__connection

    @property
    def access_address(self):
        if self.__connection is None:
            return 0x8e89bed6
        else:
            return self.__connection.access_address

    @property
    def crc_init(self):
        if self.__connection is None:
            return 0x555555
        else:
            return self.__connection.crc_init

    @property
    def hop_interval(self):
        if self.__connection is None:
            return None
        else:
            return self.__connection.hop_interval

    @property
    def hop_increment(self):
        if self.__connection is None:
            return None
        else:
            return self.__connection.hop_interval

    @property
    def channel_map(self):
        if self.__connection is None:
            return None
        else:
            return self.__connection.channel_map

    def on_synchronized(self, access_address=None, crc_init=None, hop_increment=None, hop_interval=None, channel_map=None):
        self.__synchronized = True
        self.__connection = SynchronizedConnection(
            access_address = access_address,
            crc_init = crc_init,
            hop_increment = hop_increment,
            hop_interval = hop_interval,
            channel_map = channel_map
        )
        self.trigger_event(SynchronizationEvent(self.__connection))
        logger.info("Connection synchronized -> access_address={}, crc_init={}, hop_interval={} ({} us), hop_increment={}, channel_map={}.".format(
                    "0x{:08x}".format(self.__connection.access_address),
                    "0x{:06x}".format(self.__connection.crc_init),
                    str(self.__connection.hop_interval), str(self.__connection.hop_interval*1250),
                    str(self.__connection.hop_increment),
                    "0x"+self.__connection.channel_map.hex()
        ))

    def on_desynchronized(self, access_address=None):
        self.__synchronized = False
        self.__connection = None
        self.trigger_event(DesynchronizationEvent())
        logger.info("Connection lost.")

    def _enable_sniffing(self):

        for key in self.__configuration.keys:
            self.__decryptor.add_key(key)

        if self.__configuration.access_addresses_discovery:
            self.discover_access_addresses()

        elif self.__configuration.active_connection is not None:

            access_address = self.__configuration.active_connection.access_address
            crc_init = self.__configuration.active_connection.crc_init
            channel_map = self.__configuration.active_connection.channel_map
            hop_interval = self.__configuration.active_connection.hop_interval
            hop_increment = self.__configuration.active_connection.hop_increment
            self.sniff_active_connection(access_address, crc_init, channel_map, hop_interval, hop_increment)

        elif self.__configuration.follow_connection:
            for key in self.__configuration.keys:
                self.__decryptor.add_key(key)

            if not self.can_sniff_new_connection():
                raise UnsupportedCapability("Sniff")
            else:
                self.sniff_new_connection(
                    channel=self.__configuration.channel,
                    show_advertisements=self.__configuration.show_advertisements,
                    show_empty_packets=self.__configuration.show_empty_packets,
                    bd_address=self.__configuration.filter
                )
        elif self.__configuration.show_advertisements:
            if not self.can_sniff_advertisements():
                raise UnsupportedCapability("Sniff")
            else:
                self.sniff_advertisements(
                    channel=self.__configuration.channel,
                    bd_address=self.__configuration.filter
                )


    def add_key(self, key):
        self.stop()
        self.__configuration.keys.append(key)
        self._enable_sniffing()

    def clear_keys(self):
        self.stop()
        self.__configuration.keys = []
        self._enable_sniffing()

    @property
    def decrypt(self):
        return self.__configuration.decrypt

    @decrypt.setter
    def decrypt(self, decrypt):
        self.stop()
        self.__configuration.decrypt = decrypt
        self._enable_sniffing()

    @property
    def configuration(self):
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration):
        self.stop()
        self.__configuration = new_configuration
        self._enable_sniffing()

    def configure(self, active_connection=None, access_addresses_discovery=False, advertisements=True, connection=True, empty_packets=False):
        self.stop()
        self.__configuration.active_connection = active_connection if active_connection is not None else None
        self.__configuration.access_addresses_discovery = access_addresses_discovery
        self.__configuration.show_advertisements = advertisements
        self.__configuration.show_empty_packets = empty_packets
        self.__configuration.follow_connection = connection
        self._enable_sniffing()

    @property
    def filter(self):
        return self.__configuration.filter.upper()

    @filter.setter
    def filter(self, address="FF:FF:FF:FF:FF:FF"):
        self.stop()
        self.__configuration.filter = address.upper()
        self._enable_sniffing()

    @property
    def channel(self):
        return self.__configuration.channel

    @channel.setter
    def channel(self, channel=37):
        self.stop()
        self.__configuration.channel = channel
        self._enable_sniffing()


    def available_actions(self, filter=None):
        actions = []
        if self.__synchronized:
            if self.can_inject():
                actions.append(Injector(self.device, connection=self.__connection))

            if self.can_hijack_both() or self.can_hijack_slave() or self.can_hijack_master():
                actions.append(Hijacker(self.device, connection=self.__connection))

            return [action for action in actions if filter is None or isinstance(action, filter)]

    def process_packet(self, packet):
        if self.__configuration.decrypt and BTLE_DATA in packet:
            self.__encrypted_session_initialization.process_packet(packet)
            if self.__encrypted_session_initialization.encryption:
                self.__decryptor.add_crypto_material(*self.__encrypted_session_initialization.crypto_material)
                self.__encrypted_session_initialization.reset()
            try:
                decrypted, success = self.__decryptor.attempt_to_decrypt(packet[BTLE])
                if success:
                    #packet.decrypted = decrypted
                    decrypted_packet = decrypted
                    decrypted_packet.metadata = packet.metadata
                    packet[BTLE_DATA].remove_payload()
                    packet.payload = decrypted
                    packet.metadata = decrypted_packet.metadata
                    packet.metadata.decrypted = True
                    return packet
            except MissingCryptographicMaterial:
                pass


        if self.__configuration.pairing:
            self.__legacy_pairing_cracking.process_packet(packet[BTLE])
            if self.__legacy_pairing_cracking.ready:
                keys = self.__legacy_pairing_cracking.keys
                if keys is not None:
                    tk, stk = keys
                    logger.info("[i] New temporary key extracted: ", tk.hex())
                    logger.info("[i] New short term key extracted: ", stk.hex())
                    self.trigger_event(KeyExtractedEvent(stk))
                    self.__decryptor.add_key(stk)
                    self.__legacy_pairing_cracking.reset()

        return packet

    def sniff(self):
        try:
            while True:
                if self.__configuration.access_addresses_discovery:
                    message = self.wait_for_message(filter=AccessAddressDiscovered, timeout=0.1)
                    if message is not None:
                        rssi = None
                        timestamp = None
                        if message.rssi:
                            rssi = message.rssi
                        if message.timestamp:
                            timestamp = message.timestamp
                        aa = message.access_address

                        if aa not in self.__access_addresses:
                            self.__access_addresses[aa] = AccessAddress(aa, timestamp=timestamp, rssi=rssi)
                        else:
                            self.__access_addresses[aa].update(timestamp=timestamp, rssi=rssi)
                        yield self.__access_addresses[aa]

                elif self.__configuration.active_connection is not None:
                    message = self.wait_for_message(filter=message_filter(Synchronized), timeout=0.1)

                    # TODO : improve the switch
                    if message.hop_increment > 0:
                        if self.support_raw_pdu():
                            message_type = BleRawPduReceived
                        elif self.__synchronized:
                            message_type = BlePduReceived
                        else:
                            message_type = BleAdvPduReceived

                        message = self.wait_for_message(filter=message_filter(message_type), timeout=0.1)
                        if message is not None:
                            packet = message.to_packet()
                            self.monitor_packet_rx(packet)
                            yield packet

                else:
                    if self.support_raw_pdu():
                        message_type = BleRawPduReceived
                    elif self.__synchronized:
                        message_type = BlePduReceived
                    else:
                        message_type = BleAdvPduReceived

                    message = self.wait_for_message(filter=message_filter(message_type), timeout=0.1)
                    if message is not None:
                        packet = message.to_packet()
                        packet = self.process_packet(packet)
                        self.monitor_packet_rx(packet)
                        yield packet

        # Handle device disconnection
        except WhadDeviceDisconnected:
            return



"""
Pythonic LoRaWAN 1.0 stack
"""
from binascii import unhexlify
from whad.scapy.layers.lorawan import PHYPayload
from whad.lorawan.stack.llm import LWGwLinkLayer
from whad.lorawan.helpers import EUI
from whad.common.stack import LayerState, Layer, alias, source, state

import logging
logger = logging.getLogger(__name__)

class LWGatewayState(LayerState):
    """State associated with LoRaWAN Gateway stack.
    """

    def __init__(self):
        """Initialize default state.
        """
        super().__init__()

        # Set APPKey to default value
        self.appkey = unhexlify('00000000000000000000000000000000')
        self.appeui = EUI('01:02:03:04:05:06:07:08')


@alias('phy')
@state(LWGatewayState)
class LWGatewayStack(Layer):
    """LoRaWAN Gateway stack

    This stack implements the behavior of a LoRaWAN 1.0 gateway, allowing devices to join
    through OTAA or ABP.

    For now, MAC commands are not processed.
    """

    def __init__(self, connector, options={}):
        """
        Create an instance of LoRaWAN associated with a specific connector. This
        connector provides the transport layer.

        :param connector: Connector to use with this stack.
        :type connector: WhadDeviceConnector
        :param options: Stack layer options
        :type options: dict
        """

        super().__init__(options=options)

        # APPKey must be provided
        if 'appkey' in options:
            self.state.appkey = unhexlify(options['appkey'])

        # APP EUI too
        if 'appeui' in options:
            self.state.appeui = EUI(options['appeui'])

        # Save connector (used as PHY layer)
        self.__connector = connector

        # Configure hardware to listen on a random uplink channel
        self.__connector.uplink()

    def get_appkey(self):
        """Retrieve the current APPKey

        :returns: Current APPKey
        :rtype: bytes
        """
        return self.state.appkey
    
    def get_appeui(self):
        """Retrieve the current APP EUI

        :returns: Current EUI
        :rtype: str
        """
        return self.state.appeui

    def is_device_allowed(self, dev_eui : str) -> bool:
        """Check if a device is allowed to connect.

        :param dev_eui: Device EUI
        :type dev_eui: str
        :returns: True if device is allowed to join, False otherwise.
        :rtype: bool
        """
        return self.__connector.is_device_allowed(dev_eui)

    def on_frame(self, frame : PHYPayload):
        """LoRaWAN frame callback.

        This callback handles received LoRaWAN frames.

        :param frame: LoRaWAN frame
        :type frame: PHYPayload
        """
        # Forward frame to our link layer
        self.send('ll', frame)

        # Switch back to uplink
        self.__connector.uplink()

    @source('ll')
    def send_frame(self, frame : PHYPayload, timestamp : float = None):
        """Send a LoRa frame to the current channel.

        :param frame: LoRaWAN frame to send
        :type frame: PHYPayload
        :param timestamp: Timestamp at which the payload has to be sent (if not `None`)
        :type timestamp: float
        """
        # Switch to RX1
        logger.debug('[phy] sending frame of %d bytes' % len(bytes(frame)))
        self.__connector.rx1()
        self.__connector.send(bytes(frame), timestamp=timestamp)
        
    def on_device_joined(self, dev_eui : str, dev_addr : int, appskey : bytes, nwkskey : bytes):
        """A device has just joined.

        :param dev_eui: Device EUI
        :type dev_eui: str
        :param dev_addr: Device network address
        :type dev_addr: int
        :param appskey: Device application session key
        :type appskey: bytes
        :param nwkskey: Device network encrytion session key
        :type nwkskey: bytes
        """
        # Forward notification to connector.
        self.__connector.on_device_joined(dev_eui, dev_addr, appskey, nwkskey)

    def on_data_received(self, dev_eui : str, dev_addr : int, data : bytes, upcount : int) -> bytes:
        """Received data from device.

        :param dev_eui: Device EUI
        :type dev_eui: str
        :param dev_addr: Device network address
        :type dev_addr: int
        :param data: Data received
        :type data: bytes
        :param upcount: Uplink frame counter
        :type upcount: int
        :returns: Data to send back to the device
        :rtype: bytes
        """
        return self.__connector.on_device_data(dev_eui, dev_addr, data, upcount)

LWGatewayStack.add(LWGwLinkLayer)


"""Bluetooth Low Energy connect tool

This utility will configure a compatible whad device to connect to a given
BLE device, and chain this with another tool.
"""
import sys

# Logging
import logging

from time import sleep
from queue import Queue, Empty

# Scapy
from scapy.packet import Packet

# WHAD device classes
from whad.device.unix import UnixConnector, UnixSocketServerDevice
from whad.device import Bridge

# WHAD BLE interfaces
from whad.ble import BLE, Central
from whad.hub.events import DisconnectionEvt
from whad.ble.exceptions import PeripheralNotFound

# WHAD BLE messages
from whad.hub.ble import BlePduReceived, BleRawPduReceived, Connected, Disconnected
from whad.hub.ble.bdaddr import BDAddress

# WHAD CLI helper
from whad.cli.app import CommandLineDevicePipe, run_app

logger = logging.getLogger(__name__)

class BleConnectOutputPipe(Bridge):
    """wble-connect output pipe

    When wble-connect is used with its output piped to another WHAD tool,
    it creates a connection to a BLE device and then gives access to the
    WHAD adapter in its entirety. The piped tool can send commands and will
    receive notifications from the adapter.

    There is not specific process to be applied, we are just bridging a
    unix socket server device to the central connector and that's it.
    """

    def __init__(self, input_connector, output_connector):
        """Custom output pipe, will make sure to forward queued PDUs from
        input_connector to output_connector.
        """
        # Retrieve an instance of our protocol hub
        self.hub = input_connector.hub

        # Do we get raw PDUs ?
        if input_connector.support_raw_pdu():
            self.support_raw_pdu = True
        else:
            self.support_raw_pdu = False

        # This will dissociate the underlying devices from our connectors.
        # Our central proxy will no more store queued packets, so we are left
        # with the ones we already captured.
        super().__init__(input_connector, output_connector)

        # Unlock connector, causing packets to be sent to the output connector
        input_connector.unlock(dispatch_callback=self.dispatch_pending_pdu)


    def dispatch_pending_pdu(self, pdu):
        # Convert packet back to message and forward to output
        if self.support_raw_pdu:
            message = BleRawPduReceived.from_packet(pdu)
        else:
            message = BlePduReceived.from_packet(pdu)

        # Send message to our chained tool
        self.output.send_message(message)


class BleConnectInputPipe(Bridge):
    """wble-connect input pipe

    When wble-connect has its standard input piped to another WHAD tool
    standard output, then it will consider that the previous tool configure
    another WHAD hardware adapter to get a connection and will process incoming
    connection/disconnection events as well as PDUs. These PDUs will be
    forwarded to the device wble-connect is connected to, and every PDU coming
    from our target device will be forwarded to the chained WHAD tool upstream.

    Inbound messages (sent by the device we are connected to) are monitored and
    only PDU-related messages will be forwarded  to the chained tool while
    outbound messages (coming from the chained tool) are analyzed to detect when
    a connection occurs and forward every PDU to our target device.
    """

    def __init__(self, input_connector, output_connector):
        """Initialize our wble-connect input pipe.
        """
        logger.debug('[wble-connect][output-pipe] Initialization')
        super().__init__(input_connector, output_connector)

        logger.debug('[wble-connect][output-pipe] Initialize properties')
        self.__connected = False
        self.__in_conn_handle = None
        self.__out_conn_handle = None

    def set_in_conn_handle(self, conn_handle: int):
        """Saves the input connector connection handle.
        """
        self.__in_conn_handle = conn_handle

    def set_out_conn_handle(self, conn_handle: int):
        """Saves output connection handle.
        """
        logger.debug("[wble-connect][output-pipe] set output connection handle to %d", conn_handle)
        self.__out_conn_handle = conn_handle

    def convert_packet_message(self, message, conn_handle, incoming=True):
        """Convert a BleRawPduReceived/BlePduReceived notification into the
        corresponding SendBleRawPdu/SendBlePdu command, using the provided
        connection handle.
        """
        if incoming:
            connector = self.input
        else:
            connector = self.output

        # Do we received a packet notification ?
        logger.debug("[wble-connect][output-pipe] convert message %s into a command", message)
        if isinstance(message, BleRawPduReceived):
            # Does our input connector support raw packets ?
            if connector.support_raw_pdu():
                logger.debug('[wble-connect][output-pipe] connector supports raw pdu')
                # Create a SendBleRawPdu command
                command = connector.hub.ble.create_send_raw_pdu(
                    message.direction,
                    message.pdu,
                    message.crc,
                    encrypt=False,
                    access_address=message.access_address,
                    conn_handle=conn_handle, # overwrite the connection handle
                )
                logger.debug("[wble-connect][output-pipe] created command %s", command)
            else:
                logger.debug('[wble-connect][output-pipe] connector does not support raw pdu')
                # Create a SendBlePdu command
                command = connector.hub.ble.create_send_pdu(
                    message.direction,
                    message.pdu,
                    conn_handle, # overwrite the connection handle
                    encrypt=False
                )
                logger.debug("[wble-connect][output-pipe] created command %s", command)
        elif isinstance(message, BlePduReceived):
            # Does our input connector support raw packets ?
            if connector.support_raw_pdu():
                logger.debug('[wble-connect][output-pipe] connector supports raw pdu')
                # Create a SendBleRawPdu command
                command = connector.hub.ble.create_send_raw_pdu(
                    message.direction,
                    message.pdu,
                    None,
                    encrypt=False,
                    access_address=0x11223344, # We use the default access address
                    conn_handle=conn_handle, # overwrite the connection handle
                )
                logger.debug("[wble-connect][output-pipe] created command %s", command)
            else:
                logger.debug('[wble-connect][output-pipe] connector does not support raw pdu')
                # Create a SendBlePdu command
                command = self.input.hub.ble.create_send_pdu(
                    message.direction,
                    message.pdu,
                    conn_handle, # overwrite the connection handle
                    encrypt=False
                )
                logger.debug("[wble-connect][output-pipe] created command %s", command)
        else:
            # Not a BLE packet notification
            command = None

        # Return generated command
        return command

    def on_inbound(self, message):
        """Process inbound messages.

        Inbound messages are coming from our output connector, i.e. the
        peripheral device we are connected to, we  only process any PDU-related
        message and discard other events/notifications such as connection or
        disconnection events.

        Normally, since we get packets from a central device we are supposed to
        be connected and know the connection handle corresponding to this
        connection.
        """
        if isinstance(message, BleRawPduReceived) or isinstance(message, BlePduReceived):
            if not self.__connected:
                logger.debug(
                    "[wble-connect][output-pipe] add pending inbound PDU message %s to queue",
                    message
                )
            else:
                logger.debug(
                    "[wble-connect][output-pipe] received an inbound PDU message %s",
                    message
                )
                command = self.convert_packet_message(message, self.__in_conn_handle, True)
                self.input.send_command(command)
        elif isinstance(message, Disconnected):
            # Central device has disconnected, we need to stop wble-connect.
            logger.debug("[wble-connect][output-pipe] received a disconnection notification, exiting wble-connect")

            # Detach interfaces from our bridge
            self.detach()

            # Forward the disconnection event to our central connector
            self.output.on_event(DisconnectionEvt(
                reason=message.reason,
                conn_handle=message.conn_handle
            ))

            # Close device
            self.output.close()
            self.input.close()

            # Exit process
            sys.exit(-1)
        elif isinstance(message, Connected):
            logger.debug("[wble-connect][output-pipe] received a connection notification, discard")
            return
        else:
            logger.debug("[wble-connect][output-pipe] forward default inbound message %s", message)
            # Forward other messages
            super().on_inbound(message)

    def on_outbound(self, message):
        """Process outbund messages.

        We monitor these messages to catch the connection handle used by our chained
        tool (upstream) in order to send valid commands.
        """
        if isinstance(message, BleRawPduReceived) or isinstance(message, BlePduReceived):
            if self.__out_conn_handle is not None:
                logger.debug("[wble-connect][output-pipe] received an outbound PDU message %s", message)
                command = self.convert_packet_message(message, self.__out_conn_handle, False)
                self.output.send_command(command)
        elif isinstance(message, Connected):
            # Don't forward this message.
            self.set_in_conn_handle(message.conn_handle)
            self.__connected = True
            return
        elif isinstance(message, Disconnected):
            # Chained tool has lost connection, we must handle it
            logger.debug("[wble-connect][output-pipe] received a disconnection notification, discard")
            return
        else:
            logger.debug("[wble-connect][output-pipe] forward default outbound message %s", message)
            # Forward other messages
            super().on_outbound(message)

class BleConnectApp(CommandLineDevicePipe):
    """Bluetooth Low-energy connection tool

    This command-line tool has one job: initiate a BLE connection to a specific
    device and then let another tool use it for whatever purpose it has.
    """

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD Bluetooth Low Energy connection tool',
            interface=True,
            commands=False
        )

        self.add_argument('bdaddr', metavar='BDADDR', help='Target device BD address')
        self.add_argument(
            '--spoof-public',
            metavar='PUB_BD_ADDR',
            dest='bdaddr_pub_src',
            default=None,
            help='Spoof a public BD address'
        )

        self.add_argument(
            '--spoof-random',
            metavar='RAND_BD_ADDR',
            dest='bdaddr_rand_src',
            default=None,
            help='Spoof a random BD address'
        )

        # Add an optional random type argument
        self.add_argument(
            '-r',
            '--random',
            dest='random',
            action='store_true',
            default=False,
            help='Use a random connection type'
        )


    def run(self):
        """Override App's run() method to handle scripting feature.
        """
        try:
            # Launch pre-run tasks
            self.pre_run()

            # We need to have an interface specified
            if self.interface is not None:
                # Make sure we are piped to another tool
                if self.is_stdout_piped() and not self.is_stdin_piped():
                    # Connect to the target device
                    self.connect_target(self.args.bdaddr, self.args.random)
                elif self.is_stdin_piped() and not self.is_stdout_piped():
                    # Connect to the target device once the previous tool
                    # gives us a unix socket.
                    self.connect_target_and_proxify(self.args.bdaddr, self.args.random)
                else:
                    self.error('Tool must be piped to another WHAD tool.')
            else:
                self.error('You need to specify an interface with option --interface.')

        except KeyboardInterrupt:
            self.warning("wble-connect stopped (CTL-C)")

        # Launch post-run tasks
        self.post_run()

    def set_bd_address(self, central: Central, bdaddr: str, public=True):
        """Set central BLE address
        """
        # Make sure it is a valid BD address
        if BDAddress.check(bdaddr):
            # Set the BD address
            if not central.set_bd_address(bdaddr, public=public):
                self.warning("Cannot spoof BD address, please make sure your WHAD interface supports this feature.")
        else:
            self.error(f"Invalid spoofed BD address: {bdaddr}")


    def connect_target_and_proxify(self, bdaddr, random_connection_type=False):
        """Connect to our target device and relay packets between input interface
        and our target.
        """
        # Make sure the bd address is valid
        if BDAddress.check(bdaddr):
            central = Central(self.interface)
            central.lock()

            # Spoof source BD address if required
            if self.args.bdaddr_pub_src is not None:
                self.set_bd_address(central, self.args.bdaddr_pub_src, public=True)
            elif self.args.bdaddr_rand_src is not None:
                self.set_bd_address(central, self.args.bdaddr_rand_src, public=False)

            # Connect to our target device
            try:
                central.connect(bdaddr, random_connection_type)
                output_pipe = BleConnectInputPipe(BLE(self.input_interface), central)
                output_pipe.set_out_conn_handle(central.conn_handle)

                while self.input_interface.opened:
                    sleep(.2)

            except PeripheralNotFound:
                # Could not connect
                self.error(f"Cannot connect to {bdaddr}")

    def connect_target(self, bdaddr, random_connection_type=False):
        """Connect to our target device
        """
        # Make sure the bd address is valid
        if BDAddress.check(bdaddr):
            # Configure our interface (Central with PDU queue)
            logger.debug("Using CentralProxy")
            central = Central(self.interface)
            central.lock()

            # Spoof source BD address if required
            if self.args.bdaddr_pub_src is not None:
                self.set_bd_address(central, self.args.bdaddr_pub_src, public=True)
            elif self.args.bdaddr_rand_src is not None:
                self.set_bd_address(central, self.args.bdaddr_rand_src, public=False)

            # Connect to our target device
            try:
                periph = central.connect(bdaddr, random_connection_type)

                # Get peers
                logger.info("local_peer: %s", central.local_peer)

                # Insanciate our output pipe
                proxy = UnixConnector(UnixSocketServerDevice(parameters={
                    'domain':'ble',
                    'conn_handle':periph.conn_handle,
                    'initiator_bdaddr':str(central.local_peer),
                    'initiator_addrtype':str(central.local_peer.type),
                    'target_bdaddr':str(central.target_peer),
                    'target_addrtype': str(central.target_peer.type)
                }))

                # Create our custom pipe
                BleConnectOutputPipe(central, proxy)

                # Wait for device to disconnect (or user CTL-C)
                proxy.device.wait()

            except PeripheralNotFound:
                # Could not connect
                self.error(f"Cannot connect to {bdaddr}")
            finally:
                central.stop()
        else:
            self.error(f"Invalid BD address: {bdaddr}")


def ble_connect_main():
    """Main application wrapper.
    """
    app = BleConnectApp()
    run_app(app)



"""GATT Server and Client implementation
"""
import logging

from time import time
from queue import Queue, Empty
from struct import unpack, pack
from threading import Lock

from scapy.layers.bluetooth import ATT_Handle

from whad.ble.exceptions import HookReturnValue, HookReturnAuthentRequired,\
    HookReturnAuthorRequired, HookReturnAccessDenied, HookReturnGattError, \
    HookReturnNotFound, ConnectionLostException
from whad.ble.stack.att.constants import BleAttOpcode, BleAttErrorCode, ReadAccess, \
    WriteAccess, Authentication, Authorization, Encryption
from whad.ble.stack.att.exceptions import error_response_to_exc, AttErrorCode
from whad.ble.stack.gatt.message import *
from whad.ble.stack.gatt.exceptions import GattTimeoutException
from whad.ble.profile import GenericProfile, Pairing
from whad.ble.profile.characteristic import Characteristic, CharacteristicDescriptor, ClientCharacteristicConfig, CharacteristicValue
from whad.ble.profile.service import PrimaryService, SecondaryService, IncludeService

from whad.common.stack import Layer, source, alias

logger = logging.getLogger(__name__)

def txlock(f):
    def _wrapper(self, *args, **kwargs):
        self.lock_tx()
        result = f(self, *args, **kwargs)
        self.unlock_tx()
        return result
    return _wrapper

def proclock(f):
    def _wrapper(self, *args, **kwargs):
        self.procedure_start()
        result = f(self, *args, **kwargs)
        self.procedure_stop()
        return result
    return _wrapper

@alias('gatt')
class GattLayer(Layer):
    '''Gatt client/server base class
    '''

    def configure(self, options):
        '''Configure the GATT layer
        '''
        self.__queue = Queue()
        self.__proc_lock = Lock()
        self.__tx_lock = Lock()
        self.state.terminated = False

        # Dispatch rules
        self.__handlers = {
            GattErrorResponse: self.on_error_response,

            GattFindInfoRequest: self.on_find_info_request,
            GattFindInfoResponse: self.on_find_info_response,

            GattFindByTypeValueRequest: self.on_find_by_type_value_request,
            GattFindByTypeValueResponse: self.on_find_by_type_value_response,

            GattReadByTypeRequest: self.on_read_by_type_request,
            GattReadByTypeResponse: self.on_read_by_type_response,

            GattReadByGroupTypeRequest: self.on_read_by_group_type_request,
            GattReadByGroupTypeResponse: self.on_read_by_group_type_response,

            GattReadRequest: self.on_read_request,
            GattReadResponse: self.on_read_response,

            GattReadBlobRequest: self.on_read_blob_request,
            GattReadBlobResponse: self.on_read_blob_response,

            GattReadMultipleRequest: self.on_read_multiple_request,
            GattReadMultipleResponse: self.on_read_multiple_response,

            GattWriteRequest: self.on_write_request,
            GattWriteCommand: self.on_write_command,
            GattWriteResponse: self.on_write_response,

            GattHandleValueNotification: self.on_handle_value_notification,
            GattHandleValueIndication: self.on_handle_value_indication,

            GattPrepareWriteRequest: self.on_prepare_write_request,
            GattPrepareWriteResponse: self.on_prepare_write_response,

            GattExecuteWriteRequest: self.on_execute_write_request,
            GattExecuteWriteResponse: self.on_execute_write_response,

        }

    @property
    def att(self):
        return self.get_layer('att')

    @property
    def smp(self):
        return self.get_layer('smp')

    @source('att')
    def on_att_packet(self, packet):
        '''Process ATT packets
        '''
        # Dispatch on the correct handler
        packet_type = type(packet)
        if packet_type in self.__handlers:
            self.__handlers[packet_type](packet)
        else:
            logger.error('no GATT handler for packet type %s' % packet_type.__name__)

    def procedure_start(self):
        """
        Lock GATT client or server to avoid a new procedure to be initiated while
        another is actually in progress.
        """
        logger.debug('Start of GATT procedure ...')
        self.__proc_lock.acquire()

    def procedure_stop(self):
        """Unlock GATT client or server to allow other pocedures to be executed.
        """
        logger.debug('End of GATT procedure')
        self.__proc_lock.release()

    def lock_tx(self):
        """Lock GATT client or server to process a single PDU and avoid other PDUs
        to be sent in-between.
        """
        logger.debug('Start handling PDU (tx/rx)')
        self.__tx_lock.acquire()

    def unlock_tx(self):
        """Unlock GATT client or server in order to allow other PDUs to be processed.
        """
        logger.debug('End handling PDU (tx/rx)')
        self.__tx_lock.release()

    def indicate(self, characteristic):
        """Send an indication to a GATT client. Not implemented by default.
        """
        pass

    def notify(self, characteristic):
        """Send a notfication to a GATT client. Not implemented by default.
        """
        pass

    def on_gatt_message(self, message):
        """Add a GATT message into our message queue

        :param message: GATT message to add to our queue
        """
        self.__queue.put(message, block=True, timeout=None)

    def wait_for_message(self, message_clazz, timeout=5.0):
        """Wait for a specific message type or error, other messages are dropped

        :param type message_clazz: Expected message class
        :param float timeout: Timeout value (default: 30 seconds)
        """
        start_time = time()
        while (time() - start_time) < timeout:
            # Check if connection has been terminated.
            if self.state.terminated:
                logger.debug('Connection lost')
                raise ConnectionLostException(None)
            try:
                msg = self.__queue.get(block=False,timeout=0.5)
                if isinstance(msg, message_clazz) or isinstance(msg, GattErrorResponse):
                    return msg
            except Empty:
                pass
        raise GattTimeoutException


    def error(self, request, handle, reason):
        """Send ATT Error Response
        """
        self.att.error_response(
            request,
            handle,
            reason
        )

    def on_error_response(self, error: GattErrorResponse):
        self.on_gatt_message(error)

    @txlock
    def on_find_info_request(self, packet: GattFindInfoRequest):
        '''ATT Find Information Request callback

        By default, this method generates an ATT Error Response with ATTRIBUTE_NOT_FOUND error.

        :param int start: Start handle value
        :param int end: End handle value
        '''
        logger.debug('[gatt] find_info_request, start:%d, end:%d' %(
            packet.start, packet.end
        ))
        self.error(
            BleAttOpcode.FIND_INFO_REQUEST, packet.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND
        )

    def on_find_info_response(self, packet: GattFindInfoResponse):
        '''ATT Find Information Response callback

        :param format: Information data format
        :param handles: List of handles
        '''
        pass

    @txlock
    def on_find_by_type_value_request(self, request: GattFindByTypeValueRequest):
        """ATT Find By Type Value Request callback

        :param GattFindByTypeValueRequest request: Request
        """
        logger.debug('[gatt] FindByTypeValueRequest, start: %d, type: %s' % (request.start, request.type))
        self.error(
            BleAttOpcode.FIND_BY_TYPE_VALUE_REQUEST, request.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND
        )

    def on_find_by_type_value_response(self, response: GattFindByTypeValueResponse):
        """ATT Find By Type Value Response callback

        :param GattFindByTypeValueResponse response: Response message
        """
        pass

    @txlock
    def on_read_by_type_request(self, request: GattReadByTypeRequest):
        """ATT Read By Type Request callback

        :param int start: Start handle value
        :param int end: End handle value
        :param uuid: Type UUID
        """
        self.error(
            BleAttOpcode.READ_BY_TYPE_REQUEST, request.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND
        )

    def on_read_by_type_response(self, response):
        """ATT Read By Type Response callback

        :param GattReadByTypeResponse response: Response
        """
        pass

    @txlock
    def on_read_by_group_type_request(self, request: GattReadByGroupTypeRequest):
        """ATT Read By Group Type Request callback

        :param int start: Start handle value
        :param int end: End handle value
        :param uuid: Type UUID
        """
        self.error(
            BleAttOpcode.READ_BY_GROUP_TYPE_REQUEST, request.start, BleAttErrorCode.ATTRIBUTE_NOT_FOUND
        )


    def on_read_by_group_type_response(self, response):
        """ATT Read By Group Type Response callback

        :param int item_length: Item length
        :param data: List of items
        """
        pass

    @txlock
    def on_read_request(self, request: GattReadRequest):
        """ATT Read Request callback

        :param int handle: Attribute handle
        """
        self.error(
            BleAttOpcode.READ_REQUEST, request.handle, BleAttErrorCode.INVALID_HANDLE
        )

    def on_read_response(self, response: GattReadResponse):
        """ATT Read Response callback

        :param value: Attribute value
        """
        pass

    @txlock
    def on_read_blob_request(self, request: GattReadBlobRequest):
        """ATT Read Blob Request callback

        :param int handle: Attribute handle
        :param int offset: Offset of the first byte of data to read
        """
        self.error(
            BleAttOpcode.READ_BLOB_REQUEST, request.handle, BleAttErrorCode.INVALID_HANDLE
        )

    def on_read_blob_response(self, response: GattReadBlobResponse):
        """ATT Read Blob Response callback

        :param value: Attribute value
        """
        pass

    @txlock
    def on_read_multiple_request(self, request: GattReadMultipleRequest):
        """ATT Read Multiple Request callback

        :param handles: List of handles
        """
        self.error(
            BleAttOpcode.READ_MULTIPLE_REQUEST, request.handles[0], BleAttErrorCode.INVALID_HANDLE
        )

    def on_read_multiple_response(self, response: GattReadMultipleResponse):
        """ATT Read Multiple Response callback

        :param values: Multiple Attribute values
        """
        pass


    @txlock
    def on_write_request(self, request: GattWriteRequest):
        """ATT Write Request callback

        :param int handle: Attribute handle
        :param data: Attribute value
        """
        self.error(
            BleAttOpcode.WRITE_REQUEST, request.handle, BleAttErrorCode.INVALID_HANDLE
        )

    def on_write_response(self, response):
        """ATT Write Response callback
        """
        pass

    @txlock
    def on_write_command(self, request: GattWriteCommand):
        """ATT Write Command callback

        :param int handle: Attribute handle
        :param data: Attribute data
        """
        self.error(
            BleAttOpcode.WRITE_COMMAND, request.handle, BleAttErrorCode.INVALID_HANDLE
        )

    def on_handle_value_notification(self, notification: GattHandleValueNotification):
        """ATT Handle Value Notification

        :param int handle: Attribute handle
        :param value: Attribute value
        """
        pass

    @txlock
    def on_handle_value_indication(self, indication: GattHandleValueIndication):
        """ATT Handle Value Indication

        :param int handle: Attribute handle
        :param value: Attribute value
        """
        self.att.handle_value_confirmation()

    @txlock
    def on_prepare_write_request(self, request: GattPrepareWriteRequest):
        """ATT Prepare Write request callback

        :param int handle: Attribute handle
        :param int offset: Data offset
        :param data: Attribute data
        """
        self.error(
            BleAttOpcode.PREPARE_WRITE_REQUEST, request.handle, BleAttErrorCode.INVALID_HANDLE
        )

    def on_prepare_write_response(self, response: GattPrepareWriteResponse):
        """ATT Prepare Write Response Callback

        :param int handle: Attribute handle
        :param int offset: Data offset
        :param data: Attribute data
        """
        pass

    def on_execute_write_request(self, request: GattExecuteWriteRequest):
        """ATT Execute Write Request callback

        :param int flags: Flags
        """
        pass

    def on_execute_write_response(self, response: GattExecuteWriteResponse):
        """ATT Execute Write Response callback
        """
        pass



class GattClient(GattLayer):

    def __init__(self, parent=None, layer_name=None, options={}):
        super().__init__(parent=parent, layer_name=layer_name, options=options)
        self.__model = None
        self.__notification_callbacks = {}

    def configure(self, options):
        '''Configure GATT client.
        '''
        super().configure(options)


    @property
    def model(self):
        return self.__model

    def set_model(self, model):
        '''Set the underlying GATT profile.
        '''
        if isinstance(model, GenericProfile):
            self.__model = model

    def set_client_model(self, model):
        '''This method is used with GattClientServer to specify the GATT
        client device profile.
        '''
        if isinstance(model, GenericProfile):
            self.__model = model


    ###################################
    # Supported response handlers
    ###################################

    def on_read_by_group_type_response(self, response: GattReadByGroupTypeResponse):
        """ATT Read By Group Type Response callback
        """
        self.on_gatt_message(response)

    def on_find_info_response(self, response):
        """ATT Find Information Response callback

        :param format: Information data format
        :param handles: List of handles
        """
        self.on_gatt_message(response)

    def on_read_response(self, response):
        """ATT Read Response callback

        :param value: Attribute value
        """
        self.on_gatt_message(response)

    def on_read_by_type_response(self, response: GattReadByTypeResponse):
        """ATT Read By Type Response callback

        :param GattReadByTypeResponse response: Response
        """
        self.on_gatt_message(response)

    def on_read_blob_response(self, response: GattReadBlobResponse):
        """ATT Read Blob Response callback

        :param value: Attribute value
        """
        self.on_gatt_message(response)

    def on_write_response(self, response):
        """ATT Write Response callback
        """
        self.on_gatt_message(response)

    def on_prepare_write_response(self, response):
        """ATT Prepare Write Response Callback

        :param GattPrepareWriteResponse response: Response message
        """
        self.on_gatt_message(response)

    def on_execute_write_response(self, response):
        """ATT Execute Write Response Callback

        :param GattExecuteWriteResponse response: Response message
        """
        self.on_gatt_message(response)

    def on_find_by_type_value_response(self, response: GattFindByTypeValueResponse):
        """ATT Find By Type Value Response callback

        :param GattFindByTypeValueResponse response: Response message
        """
        self.on_gatt_message(response)

    def on_handle_value_notification(self, notification):
        """ATT Handle Value Notification

        :param int handle: Attribute handle
        :param value: Attribute value
        """
        if notification.handle in self.__notification_callbacks:
            self.__notification_callbacks[notification.handle](
                notification.handle,
                notification.value,
                indication=False
            )

    @txlock
    def on_handle_value_indication(self, notification):
        """ATT Handle Value Indication

        :param int handle: Attribute handle
        :param value: Attribute value
        """
        if notification.handle in self.__notification_callbacks:
            self.__notification_callbacks[notification.handle](
                notification.handle,
                notification.value,
                indication=True
            )
        self.att.handle_value_confirmation()

    def on_exch_mtu_response(self, response):
        """ATT MTU exchange response

        :param GattExchangeMtuResponse response: response from remote device
        """
        self.on_gatt_message(response)


    ###################################
    # GATT procedures
    ###################################

    def register_notification_callback(self, handle, cb):
        self.__notification_callbacks[handle] = cb

    def unregister_notification_callback(self, handle):
        if handle in self.__notification_callbacks:
            del self.__notification_callbacks[handle]

    def clear_notification_callbacks(self):
        """Clear all notification callbacks.
        """
        self.__notification_callbacks = {}

    @proclock
    def discover_primary_service_by_uuid(self, uuid):
        """Discover a primary service by its UUID.

        :param UUID uuid: Service UUID
        :return: Service if service has been found, None otherwise
        """
        # Send FindByTypeValueRequest
        self.lock_tx()
        self.att.find_by_type_value_request(
            1,
            0xFFFF,
            0x2800,
            uuid.packed
        )
        self.unlock_tx()

        msg = self.wait_for_message(GattFindByTypeValueResponse)
        if isinstance(msg, GattFindByTypeValueResponse):
            for item in msg:
                return PrimaryService(
                    uuid=None,
                    handle=item.handle,
                    end_handle=item.end
                )
        elif isinstance(msg, GattErrorResponse):
            if msg.reason == AttErrorCode.ATTR_NOT_FOUND:
                return None
            else:
                raise error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def discover_primary_services(self):
        """Discover remote Primary Services.

        This function will yield every discovered primary service.
        """
        # List primary services handles
        handle = 1
        while True:
            # Send a Read By Group Type Request
            self.lock_tx()
            self.att.read_by_group_type_request(
                handle,
                0xFFFF,
                0x2800
            )
            self.unlock_tx()

            msg = self.wait_for_message(GattReadByGroupTypeResponse)
            if isinstance(msg, GattReadByGroupTypeResponse):
                for item in msg:
                    yield PrimaryService(
                        uuid=UUID(item.value),
                        handle=item.handle,
                        end_handle=item.end
                    )
                    handle = item.end

                    if handle == 0xFFFF:
                        return

                handle += 1
            elif isinstance(msg, GattErrorResponse):
                if msg.reason == AttErrorCode.ATTR_NOT_FOUND:
                    break
                else:
                    error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def discover_secondary_services(self):
        """Discover remote Secondary Services.
        """
        # List primary services handles
        handle = 1
        while True:
            # Send a Read By Group Type Request
            self.lock_tx()
            self.att.read_by_group_type_request(
                handle,
                0xFFFF,
                0x2801
            )
            self.unlock_tx()

            msg = self.wait_for_message(GattReadByGroupTypeResponse)
            if isinstance(msg, GattReadByGroupTypeResponse):
                for item in msg:
                    yield SecondaryService(
                        uuid=UUID(item.value),
                        handle=item.handle,
                        end_handle=item.end
                    )
                    handle = item.end

                    if handle == 0xFFFF:
                        return

                handle += 1
            elif isinstance(msg, GattErrorResponse):
                if msg.reason == AttErrorCode.ATTR_NOT_FOUND:
                    break
                else:
                    error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def discover_characteristics(self, service):
        """
        Discover service characteristics
        """
        logger.debug('discover characteristics for service %s' % service.uuid)
        logger.debug('discover characteristics from handle %d to %d' % (
            service.handle, service.end_handle
        ))
        if isinstance(service, PrimaryService):
            handle = service.handle
        else:
            return

        while handle <= service.end_handle:
            logger.debug('service end handle is: %d' % service.end_handle)
            self.lock_tx()
            self.att.read_by_type_request(
                handle,
                service.end_handle,
                0x2803
            )
            self.unlock_tx()

            msg = self.wait_for_message(GattReadByTypeResponse)
            if isinstance(msg, GattReadByTypeResponse):
                for item in msg:
                    charac_properties = item.value[0]
                    #charac_handle = unpack('<H', item.value[1:3])[0]
                    charac_handle = item.handle
                    charac_value_handle = unpack('<H', item.value[1:3])[0]
                    charac_uuid = UUID(item.value[3:])
                    charac = Characteristic(
                        uuid=charac_uuid,
                        properties=charac_properties
                    )
                    charac.handle = charac_handle
                    charac.value_handle = charac_value_handle
                    handle = charac.handle+2
                    logger.debug('found characteristic %s with handle %d' % (charac_uuid, charac_value_handle))
                    yield charac

            elif isinstance(msg, GattErrorResponse):
                if msg.reason == AttErrorCode.ATTR_NOT_FOUND:
                    break
                else:
                    error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def discover_characteristic_descriptors(self, characteristic):
        """Find characteristic descriptor
        """
        if isinstance(characteristic, Characteristic):
            handle = characteristic.value_handle + 1
            end_handle = self.__model.find_characteristic_end_handle(characteristic.handle)
            while handle <= end_handle:
                self.lock_tx()
                self.att.find_info_request(
                    handle,
                    end_handle
                )
                self.unlock_tx()

                msg = self.wait_for_message(GattFindInfoResponse)
                if isinstance(msg, GattFindInfoResponse):
                    for descriptor in msg:
                        handle = descriptor.handle

                        # End discovery if returned handle is Ending Handle (0xFFFF)
                        if handle == 0xFFFF:
                            return

                        yield(descriptor)
                elif isinstance(msg, GattErrorResponse):
                    if msg.reason == AttErrorCode.ATTR_NOT_FOUND:
                        break
                    else:
                        error_response_to_exc(msg.reason, msg.request, msg.handle)

                handle += 1

    def discover(self):
        # Discover services
        services = []
        for service in self.discover_primary_services():
            services.append(service)

        for service in services:
            for characteristic in self.discover_characteristics(service):
                service.add_characteristic(characteristic)
            self.__model.add_service(service)

        # Searching for descriptors
        for service in self.__model.services():
            for characteristic in service.characteristics():
                for descriptor in self.discover_characteristic_descriptors(characteristic):
                    if descriptor.uuid == UUID(0x2902):
                        characteristic.add_descriptor(
                            ClientCharacteristicConfig(
                                characteristic,
                                handle=descriptor.handle
                            )
                        )

    @proclock
    def read(self, handle):
        """Read a characteristic or a descriptor.

        :param int handle: Handle of the attribute to read (descriptor or characteristic)
        """
        self.lock_tx()
        self.att.read_request(gatt_handle=handle)
        self.unlock_tx()

        msg = self.wait_for_message(GattReadResponse)
        if isinstance(msg, GattReadResponse):
            return msg.value
        elif isinstance(msg, GattErrorResponse):
            raise error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def read_blob(self, handle, offset=0):
        """Read a characteristic or a descriptor starting from `offset`.

        :param int handle: Handle of the characteristic value or descriptor to read.
        :param int offset: Start reading from this offset value (default: 0)
        """
        self.lock_tx()
        self.att.read_blob_request(handle, offset)
        self.unlock_tx()

        msg = self.wait_for_message(GattReadBlobResponse)
        if isinstance(msg, GattReadBlobResponse):
            return msg.value
        elif isinstance(msg, GattErrorResponse):
            raise error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def read_long(self, handle):
        """Read a long characteristic or descriptor

        :param int handle: Handle of the attribute to read (descriptor or characteristic)
        """
        local_mtu = self.get_layer('l2cap').get_local_mtu()

        value=b''
        offset=0
        while True:
            # Send a ReadBlob request
            self.lock_tx()
            self.att.read_blob_request(handle, offset)
            self.unlock_tx()

            msg = self.wait_for_message(GattReadBlobResponse)
            if isinstance(msg, GattReadBlobResponse):
                if len(msg.value) < (local_mtu - 1):
                    value += msg.value
                    break
                else:
                    value += msg.value
                    offset += len(msg.value)
            elif isinstance(msg, GattErrorResponse):
                raise error_response_to_exc(msg.reason, msg.request, msg.handle)
        return value

    @proclock
    def write(self, handle, value):
        """Write data to a characteristic or a descriptor

        :param int handle: Target characteristic or descriptor handle
        :param bytes value: Data to write
        """
        local_mtu = self.get_layer('l2cap').get_local_mtu()

        # Check if data to write is longer than MTU
        if len(value) > (local_mtu - 3):
            # Redirect to write_long_nolock()
            return self.write_long_nolock(handle, value)
        else:
            # If data can be sent in a single write request, just send it :)
            self.lock_tx()
            self.att.write_request(
                handle,
                value
            )
            self.unlock_tx()

            msg = self.wait_for_message(GattWriteResponse)
            if isinstance(msg, GattWriteResponse):
                return True
            elif isinstance(msg, GattErrorResponse):
                raise error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def write_command(self, handle, value):
        """Write data to a characteristic or a descriptor, do not expect an answer.

        :param int handle: Target characteristic or descriptor handle
        :param bytes value: Data to write
        """
        self.lock_tx()
        self.att.write_command(
            handle,
            value
        )
        self.unlock_tx()

        # Write command does not cause the GATT server to return a response.
        return True

    @proclock
    def write_long(self, handle, value):
        """Write a long value into a characteristic

        This function wraps a call to `write_long_nolock()` to ensure this procedure
        is not started while another GATT procedure is in progress.
        """
        return self.write_long_nolock(handle, value)

    def write_long_nolock(self, handle, value):
        """Write long data (size > ATT_MTU-2) to a characteristic value.

        This method uses Prepared Write requests as described in Vol 3, Part G,
        section 4.9.5.

        :param int handle: Target characteristic value handle
        :param bytes value: Data to write
        """
        # We need to determine how many prepared write requests we should issue
        # to the target device
        data_len = len(value)
        local_mtu = self.get_layer('l2cap').get_local_mtu()

        nb_chunks = int(data_len / (local_mtu - 5))
        if nb_chunks % (local_mtu - 5) > 0:
            nb_chunks += 1
        chunk_size = local_mtu - 5

        # Send prepared write requests
        offset = 0
        for i in range(nb_chunks):
            self.lock_tx()
            self.att.prepare_write_request(handle, offset, value[offset:offset + chunk_size])
            self.unlock_tx()

            msg = self.wait_for_message(GattPrepareWriteResponse)
            if isinstance(msg, GattPrepareWriteResponse):
                if msg.value is not None:
                    offset += len(msg.value)
                else:
                    break
            elif isinstance(msg, GattErrorResponse):
                raise error_response_to_exc(msg.reason, msg.request, msg.handle)

        # Execute write request
        self.lock_tx()
        self.att.execute_write_request(1)
        self.unlock_tx()

        msg = self.wait_for_message(GattExecuteWriteResponse)
        if isinstance(msg, GattExecuteWriteResponse):
            return True
        elif isinstance(msg, GattErrorResponse):
            raise error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def read_characteristic_by_uuid(self, uuid, start=1, end=0xFFFF):
        """Read a characteristic given its UUID if its handle is comprised in a given range.

        :param UUID uuid: Characteristic UUID
        :param int start: Start handle value
        :param int end: End handle value
        """
        if uuid.type == UUID.TYPE_16:
            self.att.read_by_type_request(start, end, uuid.value())
        elif uuid.type == UUID.TYPE_128:
            # Required by scapy
            uuid1 = unpack('<Q', uuid.packed[:8])[0]
            uuid2 = unpack('<Q', uuid.packed[8:])[0]

            self.lock_tx()
            self.att.read_by_type_request_128bit(start, end, uuid1, uuid2)
            self.unlock_tx()

        msg = self.wait_for_message(GattReadByTypeResponse)
        if isinstance(msg, GattReadByTypeResponse):
            output = []
            for item in msg:
                output.append(item.value)
            if len(output) == 1:
                return output[0]
            else:
                return output
        elif isinstance(msg, GattErrorResponse):
            raise error_response_to_exc(msg.reason, msg.request, msg.handle)

    @proclock
    def set_mtu(self, mtu):
        '''Set (G)ATT client MTU (must be >= ATT_MTU, i.e. 23).

        This method is exposed in GattClient but really interact with the underlying
        ATT layer, and checks if we receive an answer from the remote device.

        :param int mtu: ATT MTU to use
        :return int: remote device MTU
        '''
        if mtu >= 23:
            self.lock_tx()
            self.att.exch_mtu_request(mtu)
            self.unlock_tx()

            msg = self.wait_for_message(GattExchangeMtuResponse)
            if isinstance(msg, GattExchangeMtuResponse):
                return msg.mtu
            elif isinstance(msg, GattErrorResponse):
                raise error_response_to_exc(msg.reason, msg.request, msg.handle)
        else:
            return None

    def services(self):
        return self.__model.services()

    def on_terminated(self):
        # Process termination.
        #super().on_terminated()

        # Remove all notification/indication callbacks
        self.clear_notification_callbacks()


class GattServer(GattLayer):
    """
    BLE GATT server
    """

    def __init__(self, parent=None, layer_name=None, options={}):
        super().__init__(parent=parent, layer_name=layer_name, options=options)
        self.__server_model = None

        # Prepared write queues
        self.__write_queues = {}

        # Subscribed characteristics
        self.__subscribed_characs = []

    @property
    def server_model(self):
        """Retrieve the current server model
        """
        return self.__server_model

    def set_model(self, model):
        '''Set the GATT server underlying profile.
        '''
        self.__server_model = model

    def set_server_model(self, model):
        '''This method is used with GattClientServer to specify the GATT
        server profile.
        '''
        # Note: redundant with set_server_model
        self.__server_model = model

    def configure(self, options):
        '''Configure GATT client.
        '''
        super().configure(options)


    ###################################
    # Supported response handlers
    ###################################


    ###################################
    # GATT procedures
    ###################################

    def notify(self, characteristic):
        """Sends a notification to a GATT client for a given characteristic.

        :param Characteristic characteristic: Characteristic to notify the GATT client about.
        """
        try:
            local_mtu = self.get_layer('l2cap').get_local_mtu()

            # Call model callback
            service = self.server_model.find_service_by_characteristic_handle(characteristic.handle)
            self.server_model.on_notification(
                service,
                characteristic,
                characteristic.value[:local_mtu-3]
            )

            # Send notification
            self.att.handle_value_notification(
                characteristic.value_handle,
                characteristic.value[:local_mtu-3]
            )
        except HookReturnValue as value_override:
            # Return overriden value
            self.att.handle_value_notification(
                characteristic.value_handle,
                value_override.value[:local_mtu-3]
            )

    @proclock
    def indicate(self, characteristic):
        """Sends an indication to a GATT client for a given characteristic.

        :param Characteristic characteristic: Characteristic to notify the GATT client about.
        """
        try:
            local_mtu = self.get_layer('l2cap').get_local_mtu()

            # Call model callback
            service = self.server_model.find_service_by_characteristic_handle(characteristic.handle)
            self.server_model.on_indication(
                service,
                characteristic,
                characteristic.value[:local_mtu-3]
            )

            # Send notification
            self.att.handle_value_indication(
                characteristic.value_handle,
                characteristic.value[:local_mtu-3]
            )
        except HookReturnValue as value_override:
            # Return overriden value
            self.att.handle_value_indication(
                characteristic.value_handle,
                value_override.value[:local_mtu-3]
            )

    @txlock
    def on_find_info_request(self, request):
        """Find information request
        """
        # List attributes by type UUID, sorted by handles
        attrs = {}
        attrs_handles = []
        for attribute in self.server_model.find_objects_by_range(request.start, request.end):
            attrs[attribute.handle] = attribute
            attrs_handles.append(attribute.handle)
        attrs_handles.sort()

        # If we have at least one item to return
        if len(attrs_handles) > 0:

            # Get MTU
            mtu = self.get_layer('l2cap').get_local_mtu()

            # Get item size (UUID size + 2)
            uuid_size = len(attrs[attrs_handles[0]].type_uuid.packed)
            if uuid_size == 2:
                item_format = 1
            else:
                item_format = 2
            item_size = uuid_size + 2
            max_nb_items = int((mtu - 2) / item_size)

            # Create our datalist
            datalist = GattAttributeDataList(item_size)

            # Iterate over items while UUID size matches and data fits in MTU
            for i in range(max_nb_items):
                if i < len(attrs_handles):
                    handle = attrs_handles[i]
                    attr_obj = attrs[handle]
                    if len(attr_obj.type_uuid.packed) == uuid_size:
                        datalist.append(
                            GattHandleUUIDItem(
                                attr_obj.handle,
                                attr_obj.type_uuid
                            )
                        )
                else:
                    break

            # Once datalist created, send answer
            datalist_raw = datalist.to_bytes()
            self.att.find_info_response(item_format, datalist_raw)
        else:
            self.error(
               BleAttOpcode.FIND_INFO_REQUEST,
               request.start,
               BleAttErrorCode.ATTRIBUTE_NOT_FOUND
            )

    @txlock
    def on_find_by_type_value_request(self, request: GattFindByTypeValueRequest):
        """ATT Find By Type Value Request callback

        :param GattFindByTypeValueRequest request: Request
        """
        # List attributes by type UUID, sorted by handles
        attrs = {}
        attrs_handles = []
        for attribute in self.server_model.find_objects_by_range(request.start, request.end):
            attrs[attribute.handle] = attribute
            attrs_handles.append(attribute.handle)
        attrs_handles.sort()
        print(attrs_handles)
        print(request.value)

        # Loop on attributes and return the attributes with a value that matches the request value
        matching_attrs = []
        for handle in attrs_handles:
            # Retrieve attribute based on handle
            attr = attrs[handle]

            # If attribute is a characteristic value or a descriptor, we make sure the characteristic
            # is readable before matching its value with the request value
            if isinstance(attr, CharacteristicValue) or isinstance(attr, CharacteristicDescriptor):
                if attr.characteristic.readable():
                    # Find characteristic end handle
                    if attr.value == request.value:
                        matching_attrs.append((handle, attr.characteristic.end_handle))
            else:
                # PrimaryService and SecondaryService are grouping types
                if isinstance(attr, PrimaryService) or isinstance(attr, SecondaryService):
                    if attr.value == request.value:
                        matching_attrs.append((handle, attr.end_handle))
                else:
                    if attr.value == request.value:
                        matching_attrs.append((handle, handle))

        # If we have found at least one attribute that matches the request, return a
        # FindByTypeValueResponse PDU
        if len(matching_attrs) > 0:
            # Build the response
            mtu = self.get_layer('l2cap').get_local_mtu()
            max_nb_items = int((mtu - 1) / 4)

            # Create our datalist
            handles_list = []

            # Iterate over items while UUID size matches and data fits in MTU
            for i in range(max_nb_items):
                if i < len(matching_attrs):
                    handle, end_handle = matching_attrs[i]
                    attr_obj = attrs[handle]
                    handles_list.append(
                        ATT_Handle(
                            handle=handle,
                            value=end_handle
                        )
                    )
                else:
                    break

            # Once datalist created, send answer
            self.att.find_by_type_value_response(handles_list)
        else:
            # Attribute not found
            self.error(
               BleAttOpcode.FIND_BY_TYPE_VALUE_REQUEST,
               request.start,
               BleAttErrorCode.ATTRIBUTE_NOT_FOUND
            )


    @txlock
    def on_read_request(self, request):
        """Read attribute value (if any)

        :param int handle: Characteristic or descriptor handle
        """
        try:
            local_mtu = self.get_layer('l2cap').get_local_mtu()

            # Search attribute by handle and send respons
            attr = self.server_model.find_object_by_handle(request.handle)

            # Ensure attribute is a readable characteristic value or a descriptor
            if isinstance(attr, CharacteristicValue):

                # Check characteristic is readable
                charac = self.server_model.find_object_by_handle(request.handle - 1)

                conn_handle = self.get_layer('l2cap').get_conn_handle()
                if charac.check_security_property(ReadAccess, Authentication):
                    print("[i] authentication required for read access !")
                    if not self.get_layer('ll').state.is_authenticated(conn_handle):
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHENT
                        )
                        return
                if charac.check_security_property(ReadAccess, Encryption):
                    print("[i] encryption required for read access !")
                    if not self.get_layer('ll').state.is_encrypted(conn_handle):
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_ENCRYPTION
                        )
                        return
                if charac.check_security_property(ReadAccess, Authorization):
                    print("[i] authorization required for read access !")
                    # TODO: not supported for now
                    self.error(
                        BleAttOpcode.READ_REQUEST,
                        request.handle,
                        BleAttErrorCode.INSUFFICIENT_AUTHOR
                    )
                    return
                if charac.readable():
                    try:
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)
                        self.server_model.on_characteristic_read(
                            service,
                            charac,
                            0,
                            local_mtu - 1
                        )

                        # Make sure the returned value matches the boundaries
                        value = charac.value[:local_mtu - 1]

                        self.att.read_response(
                            value
                        )
                    except HookReturnValue as force_value:
                        # Make sure the returned value matches the boundaries
                        value = force_value.value[:local_mtu - 1]

                        self.att.read_response(
                            value
                        )
                    except HookReturnAuthentRequired as authent_error:
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHENT
                        )
                    except HookReturnAuthorRequired as author_error:
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHOR
                        )
                    except HookReturnAccessDenied as access_denied:
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.READ_NOT_PERMITTED
                        )
                    except HookReturnNotFound as not_found:
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                    except HookReturnGattError as gatt_error:
                        self.error(
                            gatt_error.request if gatt_error.request is not None else BleAttOpcode.READ_REQUEST,
                            gatt_error.handle if gatt_error.handle is not None else request.handle,
                            gatt_error.error if gatt_error.error is not None else BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                else:
                    # Characteristic is not readable
                    self.error(
                        BleAttOpcode.READ_REQUEST,
                        request.handle,
                        BleAttErrorCode.READ_NOT_PERMITTED
                    )
            elif isinstance(attr, Characteristic):
                # Return characteristic value
                self.att.read_response(
                    attr.payload()
                )
            elif isinstance(attr, PrimaryService):
                # Return primary service value
                self.att.read_response(
                    attr.payload()
                )
            elif isinstance(attr, CharacteristicDescriptor):
                # Make sure the returned value matches the boundaries
                self.att.read_response(
                    attr.value[:local_mtu - 1]
                )

        except IndexError as e:
            self.error(
                BleAttOpcode.READ_REQUEST,
                request.handle,
                BleAttErrorCode.ATTRIBUTE_NOT_FOUND
            )

    @txlock
    def on_read_blob_request(self, request: GattReadBlobRequest):
        """Read blob request
        """
        try:
            local_mtu = self.get_layer('l2cap').get_local_mtu()

            # Search attribute by handle and send response
            attr = self.server_model.find_object_by_handle(request.handle)

            if request.offset < len(attr.value):

                # If attribute is a characteristic, make sure it is readable
                # before returning a value.
                if isinstance(attr, CharacteristicValue):
                    try:
                        charac = self.server_model.find_object_by_handle(request.handle - 1)
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                        conn_handle = self.get_layer('l2cap').get_conn_handle()
                        if charac.check_security_property(ReadAccess, Authentication):
                            print("[i] authentication required for read access !")
                            if not self.get_layer('ll').state.is_authenticated(conn_handle):
                                self.error(
                                    BleAttOpcode.READ_BLOB_REQUEST,
                                    request.handle,
                                    BleAttErrorCode.INSUFFICIENT_AUTHENT
                                )
                                return
                        if charac.check_security_property(ReadAccess, Encryption):
                            print("[i] encryption required for read access !")
                            if not self.get_layer('ll').state.is_encrypted(conn_handle):
                                self.error(
                                    BleAttOpcode.READ_BLOB_REQUEST,
                                    request.handle,
                                    BleAttErrorCode.INSUFFICIENT_ENCRYPTION
                                )
                                return
                        if charac.check_security_property(ReadAccess, Authorization):
                            print("[i] authorization required for read access !")
                            # TODO: not supported for now
                            self.error(
                                BleAttOpcode.READ_BLOB_REQUEST,
                                request.handle,
                                BleAttErrorCode.INSUFFICIENT_AUTHOR
                            )
                            return
                        if not charac.readable():
                            self.error(
                                BleAttOpcode.READ_BLOB_REQUEST,
                                request.handle,
                                BleAttErrorCode.READ_NOT_PERMITTED
                            )
                            return

                        # Call our characteristic read hook
                        self.server_model.on_characteristic_read(
                            service,
                            charac,
                            request.offset,
                            local_mtu - 1
                        )

                        # Make sure the returned value matches the boundaries
                        value = charac.value[request.offset:request.offset + local_mtu - 1]

                        self.att.read_blob_response(
                            value
                        )

                    except HookReturnValue as force_value:
                        # Make sure the returned value matches the boundaries
                        value = force_value.value[:local_mtu - 1]

                        self.att.read_blob_response(
                            value
                        )
                    except HookReturnAuthentRequired as authent_error:
                        self.error(
                            BleAttOpcode.READ_BLOB_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHENT
                        )
                    except HookReturnAuthorRequired as author_error:
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHOR
                        )
                    except HookReturnAccessDenied as access_denied:
                        self.error(
                            BleAttOpcode.READ_BLOB_REQUEST,
                            request.handle,
                            BleAttErrorCode.READ_NOT_PERMITTED
                        )
                    except HookReturnNotFound as not_found:
                        self.error(
                            BleAttOpcode.READ_BLOB_REQUEST,
                            request.handle,
                            BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                    except HookReturnGattError as gatt_error:
                        self.error(
                            gatt_error.request if gatt_error.request is not None else BleAttOpcode.READ_REQUEST,
                            gatt_error.handle if gatt_error.handle is not None else request.handle,
                            gatt_error.error if gatt_error.error is not None else BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                elif isinstance(attr, CharacteristicDescriptor):
                    # Valid offset, return data[offset:offset + MTU - 1]
                    self.att.read_blob_response(
                        attr.value[request.offset:request.offset + local_mtu - 1]
                    )
            elif request.offset == len(attr.value):
                # Special case: when offset == attribute length then return empty data
                self.att.read_blob_response(b'')
            else:
                # Invalid offset
                self.error(
                    BleAttOpcode.READ_BLOB_REQUEST,
                    request.handle,
                    BleAttErrorCode.INVALID_OFFSET
                )
        except IndexError as e:
            # Attribute not found
            self.error(
                BleAttOpcode.READ_BLOB_REQUEST,
                request.handle,
                BleAttErrorCode.ATTRIBUTE_NOT_FOUND
            )

    @txlock
    def on_write_request(self, request):
        """Write request for characteristic or descriptor value
        """
        try:
            # Retrieve attribute from model
            attr = self.server_model.find_object_by_handle(request.handle)
            if isinstance(attr, CharacteristicValue):
                # Check the corresponding characteristic is writeable
                charac = self.server_model.find_object_by_handle(request.handle - 1)

                conn_handle = self.get_layer('l2cap').get_conn_handle()
                if charac.check_security_property(WriteAccess, Authentication):
                    print("[i] authentication required for write access !")
                    if not self.get_layer('ll').state.is_authenticated(conn_handle):
                        self.error(
                            BleAttOpcode.WRITE_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHENT
                        )
                        return
                if charac.check_security_property(WriteAccess, Encryption):
                    print("[i] encryption required for write access !")
                    if not self.get_layer('ll').state.is_encrypted(conn_handle):
                        self.error(
                            BleAttOpcode.WRITE_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_ENCRYPTION
                        )
                        return
                if charac.check_security_property(WriteAccess, Authorization):
                    print("[i] authorization required for write access !")
                    # TODO: not supported for now
                    self.error(
                        BleAttOpcode.WRITE_REQUEST,
                        request.handle,
                        BleAttErrorCode.INSUFFICIENT_AUTHOR
                    )
                    return
                if charac.writeable():
                    # Retrieve corresponding service info
                    service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                    try:
                        # Trigger our write hook
                        self.server_model.on_characteristic_write(
                            service,
                            charac,
                            0,
                            request.value,
                            False
                        )

                        # Update attribute value
                        attr.value = request.value
                        self.att.write_response()

                        # Trigger our written hook (after charac has been written)
                        value =  self.server_model.on_characteristic_written(
                            service,
                            charac,
                            0,
                            attr.value,
                            True
                        )

                    except HookReturnValue as force_value:
                        # Make sure the returned value matches the boundaries
                        attr.value = force_value.value
                        self.att.write_response()

                        # Trigger our written hook (after charac has been written)
                        value =  self.server_model.on_characteristic_written(
                            service,
                            charac,
                            0,
                            attr.value,
                            True
                        )

                    except HookReturnAuthentRequired as authent_error:
                        self.error(
                            BleAttOpcode.WRITE_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHENT
                        )
                    except HookReturnAuthorRequired as author_error:
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHOR
                        )
                    except HookReturnAccessDenied as access_denied:
                        self.error(
                            BleAttOpcode.WRITE_REQUEST,
                            request.handle,
                            BleAttErrorCode.READ_NOT_PERMITTED
                        )
                    except HookReturnNotFound as not_found:
                        self.error(
                            BleAttOpcode.WRITE_REQUEST,
                            request.handle,
                            BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                    except HookReturnGattError as gatt_error:
                        self.error(
                            gatt_error.request if gatt_error.request is not None else BleAttOpcode.READ_REQUEST,
                            gatt_error.handle if gatt_error.handle is not None else request.handle,
                            gatt_error.error if gatt_error.error is not None else BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                else:
                    self.error(
                        BleAttOpcode.WRITE_REQUEST,
                        request.handle,
                        BleAttErrorCode.WRITE_NOT_PERMITTED
                    )
            elif isinstance(attr, ClientCharacteristicConfig):
                # Fixed length, make sure size <= 2.
                if len(request.value) <= 2:
                    attr.value = request.value + attr.value[len(request.value):]
                    self.att.write_response()

                    # Notify our model
                    if attr.config == 0x0001:
                        charac = attr.characteristic
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                        # Set characteristic notification callback
                        charac.set_notification_callback(self.notify)

                        self.server_model.on_characteristic_subscribed(
                            service,
                            charac,
                            notification=True
                        )
                    elif attr.config == 0x0002:
                        charac = attr.characteristic
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                        # Set characteristic indication callback
                        charac.set_indication_callback(self.indicate)

                        self.server_model.on_characteristic_subscribed(
                            service,
                            charac,
                            indication=True
                        )
                    elif attr.config == 0x0000:
                        charac = attr.characteristic
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                        # Unset characteristic indication and notification callbacks
                        charac.set_notification_callback(None)
                        charac.set_indication_callback(None)

                        # Notify model
                        self.server_model.on_characteristic_unsubscribed(
                            service,
                            charac
                        )
                else:
                    # Wrong length
                    self.error(
                        BleAttOpcode.WRITE_REQUEST,
                        request.handle,
                        BleAttErrorCode.INVALID_ATTR_VALUE_LENGTH
                    )
        except IndexError:
            self.error(
                BleAttOpcode.WRITE_REQUEST,
                request.handle,
                BleAttErrorCode.ATTRIBUTE_NOT_FOUND
            )

    @txlock
    def on_write_command(self, request):
        """Write command (without response)
        """
        try:
            # Retrieve attribute from model
            attr = self.server_model.find_object_by_handle(request.handle)
            if isinstance(attr, CharacteristicValue):
                # Check the corresponding characteristic is writeable
                charac = self.server_model.find_object_by_handle(request.handle - 1)

                conn_handle = self.get_layer('l2cap').get_conn_handle()
                if charac.check_security_property(WriteAccess, Authentication):
                    print("[i] authentication required for write access !")
                    if not self.get_layer('ll').state.is_authenticated(conn_handle):
                        self.error(
                            BleAttOpcode.WRITE_COMMAND,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHENT
                        )
                        return
                if charac.check_security_property(WriteAccess, Encryption):
                    print("[i] encryption required for write access !")
                    if not self.get_layer('ll').state.is_encrypted(conn_handle):
                        self.error(
                            BleAttOpcode.WRITE_COMMAND,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_ENCRYPTION
                        )
                        return
                if charac.check_security_property(WriteAccess, Authorization):
                    print("[i] authorization required for write access !")
                    # TODO: not supported for now
                    self.error(
                        BleAttOpcode.WRITE_COMMAND,
                        request.handle,
                        BleAttErrorCode.INSUFFICIENT_AUTHOR
                    )
                    return
                if charac.writeable():
                    # Retrieve corresponding service info
                    service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                    try:
                        # Trigger our write hook
                        value =  self.server_model.on_characteristic_write(
                            service,
                            charac,
                            0,
                            request.value,
                            True
                        )

                        # Update attribute value
                        attr.value = request.value

                        # Trigger our written hook (after charac has been written)
                        value =  self.server_model.on_characteristic_written(
                            service,
                            charac,
                            0,
                            attr.value,
                            True
                        )

                    except HookReturnValue as force_value:
                        # Make sure the returned value matches the boundaries
                        attr.value = force_value.value

                        # Trigger our written hook (after charac has been written)
                        value =  self.server_model.on_characteristic_written(
                            service,
                            charac,
                            0,
                            attr.value,
                            True
                        )

                    except HookReturnAuthentRequired as authent_error:
                        self.error(
                            BleAttOpcode.WRITE_COMMAND,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHENT
                        )
                    except HookReturnAuthorRequired as author_error:
                        self.error(
                            BleAttOpcode.READ_REQUEST,
                            request.handle,
                            BleAttErrorCode.INSUFFICIENT_AUTHOR
                        )
                    except HookReturnAccessDenied as access_denied:
                        self.error(
                            BleAttOpcode.WRITE_COMMAND,
                            request.handle,
                            BleAttErrorCode.READ_NOT_PERMITTED
                        )
                    except HookReturnNotFound as not_found:
                        self.error(
                            BleAttOpcode.WRITE_COMMAND,
                            request.handle,
                            BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                    except HookReturnGattError as gatt_error:
                        self.error(
                            gatt_error.request if gatt_error.request is not None else BleAttOpcode.READ_REQUEST,
                            gatt_error.handle if gatt_error.handle is not None else request.handle,
                            gatt_error.error if gatt_error.error is not None else BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                        )
                else:
                    self.error(
                        BleAttOpcode.WRITE_COMMAND,
                        request.handle,
                        BleAttErrorCode.WRITE_NOT_PERMITTED
                    )

            elif isinstance(attr, ClientCharacteristicConfig):
                # Fixed length, make sure size <= 2.
                if len(request.value) <= 2:
                    attr.value = request.value + attr.value[len(request.value):]

                    # Notify our model
                    if attr.config == 0x0001:
                        charac = attr.characteristic
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                        # Set characteristic notification callback
                        charac.set_notification_callback(self.notify)
                        if charac not in self.__subscribed_characs:
                            self.__subscribed_characs.append(charac)

                        self.server_model.on_characteristic_subscribed(
                            service,
                            charac,
                            notification=True
                        )
                    elif attr.config == 0x0002:
                        charac = attr.characteristic
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                        # Set characteristic indication callback
                        charac.set_indication_callback(self.indicate)
                        if charac not in self.__subscribed_characs:
                            self.__subscribed_characs.append(charac)

                        self.server_model.on_characteristic_subscribed(
                            service,
                            charac,
                            indication=True
                        )
                    elif attr.config == 0x0000:
                        charac = attr.characteristic
                        service = self.server_model.find_service_by_characteristic_handle(charac.handle)

                        # Unset characteristic indication and notification callbacks
                        charac.set_notification_callback(None)
                        charac.set_indication_callback(None)

                        if charac in self.__subscribed_characs:
                            self.__subscribed_characs.remove(charac)

                        # Notify model
                        self.server_model.on_characteristic_unsubscribed(
                            service,
                            charac
                        )
                else:
                    # Wrong length
                    self.error(
                        BleAttOpcode.WRITE_COMMAND,
                        request.handle,
                        BleAttErrorCode.INVALID_ATTR_VALUE_LENGTH
                    )
        except IndexError:
            self.error(
                BleAttOpcode.WRITE_COMMAND,
                request.handle,
                BleAttErrorCode.ATTRIBUTE_NOT_FOUND
            )

    @txlock
    def on_prepare_write_request(self, request: GattPrepareWriteRequest):
        """Prepare write request
        """
        try:
            # Retrieve attribute from model
            attr = self.server_model.find_object_by_handle(request.handle)

            # Queue request
            if request.handle not in self.__write_queues:
                self.__write_queues[request.handle] = []
            self.__write_queues[request.handle].append(request)

            # Send response
            self.att.prepare_write_response(
                request.handle,
                request.offset,
                request.value
            )

        except IndexError:
            self.error(
                BleAttOpcode.PREPARE_WRITE_REQUEST,
                request.handle,
                BleAttErrorCode.INVALID_HANDLE
            )

    @txlock
    def on_execute_write_request(self, request: GattExecuteWriteRequest):
        """Execute write request
        """
        # Clear all prepared write queues
        if request.flags == 0:
            self.__write_queues = {}
            self.att.execute_write_response()
        elif request.flags == 1:
            # Apply write requests to items
            for handle in self.__write_queues:
                try:
                    # Retrieve attribute from model
                    attr = self.server_model.find_object_by_handle(handle)

                    if isinstance(attr, CharacteristicValue):
                        # apply each update
                        for write_req in self.__write_queues[handle]:
                            attr_value = attr.value
                            if write_req.offset > len(attr_value):
                                # Clear queues
                                self.__write_queues = {}

                                # Send error
                                self.error(
                                    BleAttOpcode.EXECUTE_WRITE_REQUEST,
                                    handle,
                                    BleAttErrorCode.INVALID_OFFSET
                                )

                                # Stop now
                                return
                            else:
                                if len(attr_value) >= (write_req.offset + len(write_req.value)):
                                    attr_value = attr_value[:write_req.offset] + write_req.value + attr_value[write_req.offset + len(write_req.value):]
                                else:
                                    attr_value = attr_value[:write_req.offset] + write_req.value
                                attr.value = attr_value
                    else:
                        # Nope, only characteristic values are supported
                        pass

                except IndexError:
                    # Clear write queues
                    self.__write_queues = {}

                    # Send error
                    self.error(
                        BleAttOpcode.EXECUTE_WRITE_REQUEST,
                        request.handle,
                        BleAttErrorCode.INVALID_HANDLE
                    )

                    # Done
                    return

            # Done !
            self.att.execute_write_response()
        else:
            # Unknown flag !
            pass

    @txlock
    def on_read_by_type_request(self, request: GattReadByTypeRequest):
        """Read attribute by type request
        """
        # List attributes by type UUID, sorted by handles
        attrs = {}
        attrs_handles = []
        for attribute in self.server_model.attr_by_type_uuid(UUID(request.type), request.start, request.end):
            attrs[attribute.handle] = attribute
            attrs_handles.append(attribute.handle)
        attrs_handles.sort()

        # If we have at least one item to return
        if len(attrs_handles) > 0:

            # Get MTU
            mtu = self.get_layer('l2cap').get_local_mtu()

            # If client is looking for characteristic declaration,
            # we compute the correct item size and maximum number
            # of items we can put in a response PDU
            #
            # In the case of a characteristic declaration, UUID could
            # be 16-bit or 128-bit long so we shall only put in the same
            # answer characteristics with same size UUIDs.

            if UUID(request.type) == UUID(0x2803):

                # Get item size (UUID size + 2)
                uuid_size = len(attrs[attrs_handles[0]].uuid.packed)
                item_size = uuid_size + 5
                max_nb_items = int((mtu - 2) / item_size)

                # Create our datalist
                datalist = GattAttributeDataList(item_size)
            elif UUID(request.type) == UUID(0x2802):

                # If client is looking for included services,
                # we compute the correct item size and maximum number
                # of items we can put in a response PDU
                #
                # In the case of an included service declaration, UUID could
                # be 16-bit or 128-bit long so we shall only put in the same
                # answer characteristics with same size UUIDs.

                # Get item size
                uuid_size = len(attrs[attrs_handles[0]].uuid.packed)
                item_size = uuid_size + 6
                max_nb_items = int((mtu - 2) / item_size)

                # Create our datalist
                datalist = GattAttributeDataList(item_size)
            else:
                max_nb_items = 0
                datalist = GattAttributeDataList(0)

            # Iterate over items while UUID size matches and data fits in MTU
            for i in range(max_nb_items):
                if i < len(attrs_handles):
                    handle = attrs_handles[i]
                    attr_obj = attrs[handle]
                    if len(attr_obj.uuid.packed) == uuid_size:
                        if isinstance(attrs[handle], Characteristic):
                            datalist.append(
                                GattAttributeValueItem(
                                    handle,
                                    pack(
                                        '<BH',
                                        attr_obj.properties,
                                        attr_obj.value_handle,
                                    ) + attr_obj.uuid.packed
                                )
                            )
                        elif isinstance(attrs[handle], IncludeService):
                            if attrs[handle].service_uuid.type == UUID.TYPE_16:
                                datalist.append(
                                    GattAttributeValueItem(
                                        handle,
                                        pack(
                                            '<HH',
                                            attr_obj.service_start_handle,
                                            attr_obj.service_end_handle,
                                        ) + attr_obj.service_uuid.packed
                                    )
                                )
                            else:
                                datalist.append(
                                    GattAttributeValueItem(
                                        handle,
                                        pack(
                                            '<HH',
                                            attr_obj.service_start_handle,
                                            attr_obj.service_end_handle,
                                        )
                                    )
                                )
                else:
                    break

            # Check that our result datalist does contain something
            if len(datalist) > 0:
                # Once datalist created, send answer
                datalist_raw = datalist.to_bytes()
                self.att.read_by_type_response(item_size, datalist_raw)
            else:
                # If not, send an error.
                self.error(
                    BleAttOpcode.READ_BY_TYPE_REQUEST,
                    request.start,
                    BleAttErrorCode.ATTRIBUTE_NOT_FOUND
                )
        else:
            self.error(
               BleAttOpcode.READ_BY_TYPE_REQUEST,
               request.start,
               BleAttErrorCode.ATTRIBUTE_NOT_FOUND
            )

    @txlock
    def on_read_by_group_type_request(self, request: GattReadByGroupTypeRequest):
        """Read by group type request

        List attribute with given type UUID from `start` handle to ̀`end` handle.
        """
        # List attributes by type UUID, sorted by handles
        attrs = {}
        attrs_handles = []
        for attribute in self.server_model.attr_by_type_uuid(UUID(request.type), request.start, request.end):
            attrs[attribute.handle] = attribute
            attrs_handles.append(attribute.handle)
        attrs_handles.sort()

        # If we have at least one item to return
        if len(attrs_handles) > 0:

            # Get MTU
            mtu = self.get_layer('l2cap').get_local_mtu()

            # Get item size (UUID size + 4)
            uuid_size = len(attrs[attrs_handles[0]].uuid.packed)
            item_size = uuid_size + 4
            max_nb_items = int((mtu - 2) / item_size)

            # Create our datalist
            datalist = GattAttributeDataList(item_size)

            # Iterate over items while UUID size matches and data fits in MTU
            for i in range(max_nb_items):
                if i < len(attrs_handles):
                    handle = attrs_handles[i]
                    end_handle = attrs[handle].end_handle
                    attr_uuid = attrs[handle].uuid
                    if len(attr_uuid.packed) == uuid_size:
                        datalist.append(
                            GattGroupTypeItem(handle, end_handle, attr_uuid.packed)
                        )
                else:
                    break

            # Once datalist created, send answer
            datalist_raw = datalist.to_bytes()
            self.att.read_by_group_type_response(item_size, datalist_raw)
        else:
            self.error(
               BleAttOpcode.READ_BY_GROUP_TYPE_REQUEST,
               request.start,
               BleAttErrorCode.ATTRIBUTE_NOT_FOUND
            )

    def on_terminated(self):
        """Connection has been terminated, remove characteristics subscriptions.
        """
        # Unsubscribe from everything and remove callbacks again
        for charac in self.__subscribed_characs:
            charac.set_notification_callback(None)
            charac.set_indication_callback(None)
        self.__subscribed_characs = []

    def pairing(self, pairing=None):
        pairing_parameters = Pairing()
        if pairing is None:
            pairing_parameters = self.server_model.get_pairing_parameters()
        elif isinstance(pairing, Pairing):
            pairing_parameters = pairing

        self.smp.request_pairing(pairing=pairing_parameters)

class GattClientServer(GattServer, GattClient):

    def __init__(self, parent=None, layer_name=None, options={}):
        super().__init__(parent=parent, layer_name=layer_name, options=options)



import json

from whad.ble.connector import BLE, Central, Peripheral
from whad.ble.profile import GenericProfile
from whad.ble.exceptions import ConnectionLostException
from whad.ble import UnsupportedCapability, message_filter

from whad.hub.events import ConnectionEvt
from whad.hub.ble import Hijacked


class Hijacker(BLE):

    def __init__(self, device, connection=None):
        super().__init__(device)
        self.__connection = connection
        self.__hijack_master = False
        self.__hijack_slave = False
        self.__status = False
        self.__exception = None
        # Check if device accepts hijacking
        if not self.can_hijack_slave() and not self.can_hijack_master():
            raise UnsupportedCapability("Hijack")

    @property
    def central(self):
        available_actions = self.available_actions(Central)
        if len(available_actions) == 1:
            return available_actions[0]
        return None


    @property
    def peripheral(self):
        available_actions = self.available_actions(Peripheral)
        if len(available_actions) == 1:
            return available_actions[0]
        return None

    def available_actions(self, filter=None):
        actions = []
        if self.__status:
            # It should be replaced by arguments in function calls, building a pseudo packet seems dirty
            if self.__hijack_master:
                pseudo_connection = ConnectionEvt()
                pseudo_connection.conn_handle = 0
                pseudo_connection.initiator = b"\x00\x00\x00\x00\x00\x00"
                pseudo_connection.init_addr_type = 0
                pseudo_connection.advertiser = b"\x00\x00\x00\x00\x00\x00"
                pseudo_connection.adv_addr_type = 0
                actions.append(Central(self.device, existing_connection=pseudo_connection))
            if self.__hijack_slave:
                pseudo_connection = ConnectionEvt()
                pseudo_connection.conn_handle = 1
                pseudo_connection.initiator = b"\x00\x00\x00\x00\x00\x00"
                pseudo_connection.init_addr_type = 0
                pseudo_connection.advertiser = b"\x00\x00\x00\x00\x00\x00"
                pseudo_connection.adv_addr_type = 0

                actions.append(
                    Peripheral(
                        self.device,
                        existing_connection = pseudo_connection
                    )
                )
        return [action for action in actions if filter is None or isinstance(action, filter)]


    def hijack(self, master = True, slave = False):
        """
        Hijack master, slave, or both
        """
        if self.__connection is not None:
            if master and slave:
                self.__hijack_master = master
                self.__hijack_slave = slave
                self.hijack_both(self.__connection.access_address)
            elif master:
                self.__hijack_master = master
                self.hijack_master(self.__connection.access_address)

            elif slave:
                self.__hijack_slave = slave
                self.hijack_slave(self.__connection.access_address)

            message = self.wait_for_message(filter=message_filter(Hijacked))
            self.__status = message.success
            return (message.success)
        else:
            raise self.__exception

    def on_desynchronized(self, access_address):
        if access_address == self.__connection.access_address:
            self.__exception = ConnectionLostException(self.__connection)
            self.__connection = None



"""
BLE GATT Service Model
======================
"""
import logging
from struct import pack

from whad.ble.exceptions import InvalidHandleValueException
from whad.ble.profile.attribute import Attribute, UUID, get_uuid_alias
from whad.ble.profile.characteristic import Characteristic

logger = logging.getLogger(__name__)

class Service(Attribute):
    def __init__(self, uuid, type_uuid, handle=0, end_handle=0):
        super().__init__(uuid=type_uuid,handle=handle, value=uuid.to_bytes())
        self.__service_uuid = uuid
        if handle > 0:
            self.__end_handle = end_handle if end_handle > 0 else handle
        else:
            self.__end_handle = 0
        self.__characteristics = []
        self.__included_services = []

    @property
    def uuid(self):
        return self.__service_uuid


    @Attribute.handle.setter
    def handle(self, new_handle):
        """Overwrite `Attribute` handle setter.
        """
        if isinstance(new_handle, int):
            # Update service handle
            Attribute.handle.fset(self, new_handle)

            # Update the underlying characteristics
            char_handle = self.handle
            for characteristic in self.__characteristics:
                characteristic.handle = char_handle + 1
                char_handle = characteristic.end_handle
            
            # Update service end_handle value
            self.__end_handle = char_handle
        else:
            raise InvalidHandleValueException

    @property
    def end_handle(self):
        return self.__end_handle

    @end_handle.setter
    def end_handle(self, value):
        self.__end_handle = value

    @property
    def name(self):
        alias = get_uuid_alias(self.__service_uuid)
        if alias is not None:
            return '%s (0x%s)' % (
                alias,
                str(self.__service_uuid)
            )
        else:
            return str(self.__service_uuid)

    def payload(self):
        """Return service UUID as bytes
        """
        return self.__service_uuid.packed

    def add_characteristic(self, characteristic):
        """Add characteristic, update end handle
        """
        if characteristic.handle == 0:
            # Add characteristic, set characteristic handle and update end handle if required
            characteristic.handle = self.end_handle + 1
            self.__end_handle = characteristic.end_handle
        elif characteristic.handle > self.__end_handle:
            # Add characteristic and update end handle if required
            self.__end_handle = characteristic.end_handle
        self.__characteristics.append(characteristic)

    def remove_characteristic(self, characteristic):
        """Remove a specific characteristic
        """
        if isinstance(characteristic, UUID):
            # Look for characteristic and remove it if found
            for charac in self.__characteristics:
                if charac.uuid == characteristic:
                    self.__characteristics.remove(charac)
                    break
        elif isinstance(characteristic, Characteristic):
            # Look for characteristic object
            if characteristic in self.__characteristics:
                self.__characteristics.remove(characteristic)
        
        # Update characteristic handles
        char_handle = self.handle
        for characteristic in self.__characteristics:
            characteristic.handle = char_handle + 1
            char_handle = characteristic.end_handle
        
        # Update service end_handle value
        self.__end_handle = char_handle

    def characteristics(self):
        for charac in self.__characteristics:
            yield charac

    def get_characteristic(self, uuid):
        """Get characteristic by UUID
        """
        for charac in self.__characteristics:
            if charac.uuid == uuid:
                return charac
        return None
    
    def add_include_service(self, included_service):
        """Add include service definition, update end handle
        """
        if included_service.handle == 0:
            # Add characteristic, set characteristic handle and update end handle if required
            included_service.handle = self.end_handle + 1
            self.__end_handle = included_service.end_handle
        elif included_service.handle > self.__end_handle:
            # Add characteristic and update end handle if required
            self.__end_handle = included_service.end_handle
        self.__included_services.append(included_service)

    def remove_include_service(self, included_service):
        """Remove a specific characteristic
        """
        if isinstance(included_service, UUID):
            # Look for characteristic and remove it if found
            for inc_service in self.__included_services:
                if inc_service.uuid == included_service:
                    self.__included_services.remove(inc_service)
                    break
        elif isinstance(included_service, IncludeService):
            # Look for characteristic object
            if included_service in self.__included_services:
                self.__included_services.remove(included_service)
        
        # Update included services and characteristic handles
        char_handle = self.handle
        for inc_service in self.__included_services:
            inc_service.handle = char_handle + 1
            char_handle = inc_service.handle
        for characteristic in self.__characteristics:
            characteristic.handle = char_handle + 1
            char_handle = characteristic.end_handle
        
        # Update service end_handle value
        self.__end_handle = char_handle        

    def included_services(self):
        for inc_service in self.__included_services:
            yield inc_service


class PrimaryService(Service):

    def __init__(self, uuid, handle=0, end_handle=0):
        super().__init__(uuid, UUID(0x2800),handle=handle, end_handle=end_handle)

class SecondaryService(Service):

    def __init__(self, uuid, handle=None):
        super().__init__(uuid, UUID(0x2801),handle=handle)

class IncludeService(Attribute):
    """IncludeService Attribute class

    This class stores the information related to an included service:

    - the included service UUID (16-bit or 128-bit UUID)
    - the start and end handles of the said included service
    """

    def __init__(self, uuid, handle=0, start_handle=0, end_handle=0):
        """Initialize an included service

        :param  uuid:       Included service UUID
        :type   uuid:       UUID
        :param  handle:     Included service start handle
        :type   handle:     int
        :param  end_handle: Included service end handle
        :type   end_handle: int
        """
        self.__service_uuid = uuid
        self.__start_handle = start_handle
        self.__end_handle = end_handle
        super().__init__(UUID(0x2802), handle=handle, value=self.payload())

    @property
    def end_handle(self):
        """Return this attribute end handle

        This attribute does not belong to a group so its end handle is the same
        than its handle.
        """
        return self.handle

    @end_handle.setter
    def end_handle(self, value):
        self.__end_handle = value

    @property
    def uuid(self):
        """Return the attribute type UUID.
        """
        return self.type_uuid
    
    @property
    def service_uuid(self):
        """Return the included service UUID
        """
        return self.__service_uuid
    
    @property
    def service_start_handle(self):
        """Return the included service start handle
        """
        return self.__start_handle
    
    @service_start_handle.setter
    def service_start_handle(self, value):
        self.__start_handle = value
    
    @property
    def service_end_handle(self):
        """Return the included service end handle
        """
        return self.__end_handle
    
    @service_end_handle.setter
    def service_end_handle(self, value):
        self.__end_handle = value

    @property
    def name(self):
        """Generate the description of the included service definition attribute.
        """
        alias = get_uuid_alias(self.__service_uuid)
        if alias is not None:
            return 'Included service %s (0x%s)' % (
                alias,
                str(self.__service_uuid)
            )
        else:
            return 'Included service ' + str(self.__service_uuid)

    def payload(self):
        """Return service UUID as bytes
        """
        if self.__service_uuid.type == UUID.TYPE_16:
            return pack('<HH', self.__start_handle, self.__end_handle) + self.__service_uuid.packed
        else:
            return pack('<HH', self.__start_handle, self.__end_handle)



"""
WHAD traffic analyzer base module.
"""
from typing import List
from scapy.packet import Packet


class TrafficAnalyzer:
    """
    Traffic analyzer base class.

    This class must be inherited by specialized traffic analyzers
    in order to provide `wanalyze` with extracted and/or computed
    data. 
    """

    def __init__(self):
        self.__triggered = False
        self.__completed = False
        self.__marked_packets = []

    def process_packet(self, packet):
        """Process a packet.
        """

    def mark_packet(self, packet):
        """Mark a specific packet.
        """
        self.__marked_packets.append(packet)

    def reset(self):
        """Reset traffic analyzer state.
        """
        self.__triggered = False
        self.__completed = False
        self.__marked_packets = []

    def trigger(self):
        """Trigger this traffic analyzer.
        """
        self.__triggered = True

    def complete(self):
        """Mark this analyzer as completed.

        Once a traffic analyzer is completed, its output
        can be queried.
        """
        self.__completed = True

    @property
    def marked_packets(self) -> List[Packet]:
        """Returns marked packets.
        """
        return self.__marked_packets

    @property
    def output(self):
        """Returns the traffic analyzer output.
        """
        return None

    @property
    def triggered(self) -> bool:
        """Determine if the traffic analyzer has been triggered.

        :return-type: bool
        :return: `True` if triggered, `False` otherwise.
        """
        return self.__triggered

    @property
    def completed(self) -> bool:
        """Determine if the traffic analyzer has completed is
        job.

        :return-type: bool
        :return: `True` if completed, `False` otherwise.
        """
        return self.__completed



"""
BLE ATT layer manager

This layer manager mostly translates Scapy ATT packets into GATT messages required by the GATT
layer, and exposes an interface to the GATT layer in order to allow ATT packets to be forged
and sent to the underlying layer (L2CAP).
"""
from typing import List
from scapy.layers.bluetooth import ATT_Error_Response, ATT_Exchange_MTU_Request, \
    ATT_Exchange_MTU_Response, ATT_Execute_Write_Request, ATT_Execute_Write_Response, \
    ATT_Find_By_Type_Value_Request, ATT_Find_By_Type_Value_Response, ATT_Find_Information_Request, \
    ATT_Find_Information_Response, ATT_Prepare_Write_Request, ATT_Prepare_Write_Response, \
    ATT_Read_Blob_Request, ATT_Handle_Value_Indication, ATT_Handle_Value_Notification, \
    ATT_Read_Blob_Response, ATT_Read_By_Group_Type_Request, ATT_Read_By_Group_Type_Response, \
    ATT_Read_By_Type_Request, ATT_Read_By_Type_Response, ATT_Read_Multiple_Request, ATT_Read_Multiple_Response, \
    ATT_Read_Request, ATT_Read_Response, ATT_Write_Command, ATT_Write_Response, ATT_Write_Request, \
    ATT_Read_By_Type_Request_128bit, ATT_Hdr, ATT_Handle
from scapy.packet import bind_layers, Packet

from whad.ble.stack.att.constants import BleAttOpcode, SecurityProperty, SecurityAccess

from whad.common.stack import Layer, alias, source, instance

#from whad.ble.stack.gatt import  Gatt
from whad.ble.stack.gatt.message import GattExecuteWriteRequest, GattExecuteWriteResponse, \
    GattFindInfoResponse, GattHandleValueIndication, GattHandleValueNotification, GattPrepareWriteRequest, \
    GattPrepareWriteResponse, GattReadByGroupTypeResponse, GattErrorResponse, \
    GattFindByTypeValueRequest, GattFindByTypeValueResponse, GattFindInfoRequest, GattReadByTypeResponse, \
    GattReadRequest, GattReadResponse, GattReadBlobRequest, GattReadBlobResponse, GattReadMultipleRequest, \
    GattReadMultipleResponse, GattWriteCommand, GattWriteRequest, GattWriteResponse, GattPrepareWriteRequest, \
    GattPrepareWriteResponse, GattExecuteWriteRequest, GattExecuteWriteResponse, \
    GattExchangeMtuResponse, GattReadByGroupTypeRequest, GattReadByTypeRequest, GattReadByTypeRequest128

# Add missing ATT_Handle_Value_Confirmation class
class ATT_Handle_Value_Confirmation(Packet):
    name = "Handle Value Confirmation"

# Bind ATT_Handle_Value_Confirmation with ATT_Hdr
bind_layers(ATT_Hdr, ATT_Handle_Value_Confirmation, opcode=BleAttOpcode.HANDLE_VALUE_CONFIRMATION)


@alias('att')
class ATTLayer(Layer):

    ##########################################
    # Incoming requests and responses
    ##########################################

    @instance('l2cap')
    def on_packet(self, instance, att_pkt):
        """Dispatch ATT packet.
        """
        if ATT_Error_Response in att_pkt:
            self.on_error_response(att_pkt.getlayer(ATT_Error_Response))
        elif ATT_Exchange_MTU_Request in att_pkt:
            self.on_exch_mtu_request(att_pkt.getlayer(ATT_Exchange_MTU_Request))
        elif ATT_Exchange_MTU_Response in att_pkt:
            self.on_exch_mtu_response(att_pkt.getlayer(ATT_Exchange_MTU_Response))

        elif ATT_Find_Information_Request in att_pkt:
            self.on_find_info_request(att_pkt.getlayer(ATT_Find_Information_Request))
        elif ATT_Find_Information_Response in att_pkt:
            self.on_find_info_response(att_pkt.getlayer(ATT_Find_Information_Response))
        elif ATT_Find_By_Type_Value_Request in att_pkt:
            self.on_find_by_type_value_request(att_pkt.getlayer(ATT_Find_By_Type_Value_Request))
        elif ATT_Find_By_Type_Value_Response in att_pkt:
            self.on_find_by_type_value_response(att_pkt.getlayer(ATT_Find_By_Type_Value_Response))
        elif ATT_Read_By_Type_Request in att_pkt:
            self.on_read_by_type_request(att_pkt.getlayer(ATT_Read_By_Type_Request))
        elif ATT_Read_By_Type_Response in att_pkt:
            self.on_read_by_type_response(att_pkt.getlayer(ATT_Read_By_Type_Response))
        elif ATT_Read_By_Group_Type_Request in att_pkt:
            self.on_read_by_group_type_request(att_pkt.getlayer(ATT_Read_By_Group_Type_Request))
        elif ATT_Read_By_Group_Type_Response in att_pkt:
            self.on_read_by_group_type_response(att_pkt.getlayer(ATT_Read_By_Group_Type_Response))
        elif ATT_Read_Request in att_pkt:
            self.on_read_request(att_pkt.getlayer(ATT_Read_Request))
        elif ATT_Read_By_Type_Request_128bit in att_pkt:
            self.on_read_by_type_request_128bit(att_pkt.getlayer(ATT_Read_By_Type_Request_128bit))
        elif ATT_Read_Response in att_pkt:
            self.on_read_response(att_pkt.getlayer(ATT_Read_Response))
        elif ATT_Read_Blob_Request in att_pkt:
            self.on_read_blob_request(att_pkt.getlayer(ATT_Read_Blob_Request))
        elif ATT_Read_Blob_Response in att_pkt:
            self.on_read_blob_response(att_pkt.getlayer(ATT_Read_Blob_Response))
        elif ATT_Read_Multiple_Request in att_pkt:
            self.on_read_multiple_request(att_pkt.getlayer(ATT_Read_Multiple_Request))
        elif ATT_Read_Multiple_Response in att_pkt:
            self.on_read_multiple_response(att_pkt.getlayer(ATT_Read_Multiple_Response))
        elif ATT_Write_Request in att_pkt:
            self.on_write_request(att_pkt.getlayer(ATT_Write_Request))
        elif ATT_Write_Response in att_pkt:
            self.on_write_response(att_pkt.getlayer(ATT_Write_Response))
        elif ATT_Write_Command in att_pkt:
            self.on_write_command(att_pkt.getlayer(ATT_Write_Command))
        elif ATT_Handle_Value_Notification in att_pkt:
            self.on_handle_value_notification(att_pkt.getlayer(ATT_Handle_Value_Notification))
        elif ATT_Handle_Value_Indication in att_pkt:
            self.on_handle_value_indication(att_pkt.getlayer(ATT_Handle_Value_Indication))
        elif ATT_Prepare_Write_Request in att_pkt:
            self.on_prepare_write_request(att_pkt.getlayer(ATT_Prepare_Write_Request))
        elif ATT_Prepare_Write_Response in att_pkt:
            self.on_prepare_write_response(att_pkt.getlayer(ATT_Prepare_Write_Response))
        elif ATT_Execute_Write_Request in att_pkt:
            self.on_execute_write_request(att_pkt.getlayer(ATT_Execute_Write_Request))
        elif ATT_Execute_Write_Response in att_pkt:
            self.on_execute_write_response(att_pkt.getlayer(ATT_Execute_Write_Response))
        # Signed command not supported yet
        # Write Response has no body
        elif att_pkt.opcode == BleAttOpcode.WRITE_RESPONSE:
            self.on_write_response(None)
        # Read Blob Response has no body
        elif att_pkt.opcode == BleAttOpcode.READ_BLOB_RESPONSE:
            self.on_read_blob_response(None)
        # Read Response has no body
        elif att_pkt.opcode == BleAttOpcode.READ_RESPONSE:
            self.on_read_response(None)
        # Execute write request
        elif att_pkt.opcode == BleAttOpcode.EXECUTE_WRITE_RESPONSE:
            self.on_execute_write_response(None)

    def on_error_response(self, error_resp):
        # Send a GattErrorResponse message to gatt
        self.send(
            'gatt',
            GattErrorResponse(
                error_resp.request,
                error_resp.handle,
                error_resp.ecode
            ),
            tag='GATT_ERROR_RSP',
        )

    def on_exch_mtu_request(self, mtu_req):
        """Handle ATT Exchange MTU request, update L2CAP TX MTU and returns
        our MTU.

        :param ATT_Exchange_MTU_Request mtu_req: MTU request
        """
        # Update L2CAP Client MTU
        self.get_layer('l2cap').set_remote_mtu(mtu_req.mtu)

        # Send back our MTU.
        self.send_data(ATT_Exchange_MTU_Response(
            mtu=self.get_layer('l2cap').get_local_mtu()
        ))

    def on_exch_mtu_response(self, mtu_resp):
        """Update L2CAP remote MTU based on ATT_Exchange_MTU_Response.

        :param mtu_resp ATT_Exchange_MTU_Response: MTU response
        """
        self.get_layer('l2cap').set_remote_mtu(mtu_resp.mtu)

        # Forward to GATT
        #self.send('gatt', GattExchangeMtuResponse(mtu=mtu_resp.mtu),
        #          tag='XCHG_MTU_RESP')


    def on_find_info_request(self, request):
        """Handle ATT Find Information Request

        :param ATT_Find_Information_Request request: Request
        """
        self.send('gatt', GattFindInfoRequest(
                request.start,
                request.end
            ), tag='FIND_INFO_REQ'
        )

    def on_find_info_response(self, response):
        """Handle ATT Find Information Response
        """
        handles = b''.join([item.build() for item in response.handles])
        self.send('gatt', GattFindInfoResponse.from_bytes(
                response.format,
                handles
            ), tag='FIND_INFO_RESP',
        )

    def on_find_by_type_value_request(self, request):
        self.send('gatt', GattFindByTypeValueRequest(
                request.start,
                request.end,
                request.uuid,
                request.data
            )
        )

    def on_find_by_type_value_response(self, response):
        handles = b''.join([item.build() for item in response.handles])
        self.send('gatt', GattFindByTypeValueResponse.from_bytes(handles))

    def on_read_by_type_request(self, request):
        """Handle read by type request
        """
        self.send('gatt', GattReadByTypeRequest(
            request.start,
            request.end,
            request.uuid
        ))

    def on_read_by_type_request_128bit(self, request):
        """Handle ATT Read By Type Request 128-bit UUID
        """
        self.send('gatt',
            request.start,
            request.end,
            request.uuid1,
            request.uuid2
        )

    def on_read_by_type_response(self, response: GattReadByTypeResponse):
        """Handle read by type response
        """
        # Must rebuild handles payload as bytes, since scapy parsed it :(
        handles = b''.join([item.build() for item in response.handles])
        self.send('gatt', GattReadByTypeResponse.from_bytes(
                response.len,
                handles
            )
        )

    def on_read_request(self, request):
        """Handle ATT Read Request
        """
        self.send('gatt', GattReadRequest(
                request.gatt_handle
            )
        )

    def on_read_response(self, response):
        """Handle ATT Read Response
        """
        if response is not None:
            self.send('gatt', GattReadResponse(
                    response.value
                )
            )
        else:
            self.send('gatt', GattReadResponse(
                    b''
                )
            )

    def on_read_blob_request(self, request):
        """Handle ATT Read Blob Request
        """
        self.send('gatt', GattReadBlobRequest(
                request.gatt_handle,
                request.offset
            )
        )

    def on_read_blob_response(self, response):
        """Handle ATT Read Blob Response

        :param response: ATT response if provided, None otherwise.
        """
        if response is not None:
            self.send('gatt', GattReadBlobResponse(
                    response.value
                )
            )
        else:
            self.send('gatt', GattReadBlobResponse(
                    None
                )
            )

    def on_read_multiple_request(self, request):
        """Handle ATT Read Multiple Request
        """
        self.send('gatt',GattReadMultipleRequest(request.handles))

    def on_read_multiple_response(self, response):
        """Handle ATT Read Multiple Response
        """
        self.send('gatt', GattReadMultipleResponse(response.values))

    def on_read_by_group_type_request(self, request):
        """Handle ATT Read By Group Type Request
        """
        self.send('gatt', GattReadByGroupTypeRequest(
                request.start,
                request.end,
                request.uuid
            )
        )

    def on_read_by_group_type_response(self, response):
        """Handle ATT Read By Group Type Response
        """
        self.send('gatt', GattReadByGroupTypeResponse.from_bytes(
                response.length,
                response.data
            )
        )

    def on_write_request(self, request):
        """Handle ATT Write Request
        """
        self.send('gatt', GattWriteRequest(
                request.gatt_handle,
                request.data
            )
        )

    def on_write_response(self, response):
        """Handle ATT Write Response
        """
        self.send('gatt', GattWriteResponse())

    def on_write_command(self, command):
        """Handle ATT Write Command
        """
        self.send('gatt', GattWriteCommand(
                command.gatt_handle,
                command.data
            )
        )

    def on_prepare_write_request(self, request):
        """Handle ATT Prepare Write Request
        """
        self.send('gatt', GattPrepareWriteRequest(
                request.gatt_handle,
                request.offset,
                request.data
            )
        )

    def on_prepare_write_response(self, response):
        """Handle ATT Prepare Write Response
        """
        self.send('gatt', GattPrepareWriteResponse(
                response.gatt_handle,
                response.offset,
                response.data
            )
        )

    def on_execute_write_request(self, request):
        """Handle ATT Execute Write Request
        """
        self.send('gatt', GattExecuteWriteRequest(
                request.flags
            )
        )

    def on_execute_write_response(self, response):
        """Handle ATT Execute Write Response
        """
        self.send('gatt', GattExecuteWriteResponse())

    def on_handle_value_notification(self, notif):
        """Handle ATT Handle Value Notification
        """
        self.send('gatt', GattHandleValueNotification(
                notif.gatt_handle,
                notif.value
            )
        )

    def on_handle_value_indication(self, notif):
        """Handle ATT Handle Value indication
        """
        self.send('gatt', GattHandleValueIndication(
                notif.gatt_handle,
                notif.value
            )
        )

    ##########################################
    # Outgoing requests and responses
    ##########################################

    def send_data(self, packet):
        self.send('l2cap', ATT_Hdr()/packet)

    def error_response(self, request, handle, reason):
        """Sends an ATT Error Response

        :param int request: Request that generated this error
        :param int handle: Attribute handle that generated this error
        :param int ecode: Reason why this error has been generated
        """
        self.send_data(ATT_Error_Response(
                request=request,
                handle=handle,
                ecode=reason
            )
        )

    def exch_mtu_request(self, mtu):
        """Sends an ATT Exchange MTU Request

        :param int mtu: Maximum Transmission Unit
        """
        # Update local MTU first
        self.get_layer('l2cap').set_local_mtu(mtu)
        self.send_data(ATT_Exchange_MTU_Request(
            mtu=mtu
        ))

    def exch_mtu_response(self, mtu):
        """Sends an ATT Exchange MTU Response

        :param int mtu: Maximum Transmission Unit
        """
        self.send_data(ATT_Exchange_MTU_Response(
            mtu=mtu
        ))


    def find_info_request(self, start, end):
        """Sends an ATT Find Information Request
        """
        self.send_data(ATT_Find_Information_Request(
            start=start,
            end=end
        ))

    def  find_info_response(self, format, handles):
        """Sends an ATT Find Information Response
        """

        self.send_data(ATT_Find_Information_Response(
            format=format,
            handles=handles
        ))

    def find_by_type_value_request(self, start, end, type_uuid, value):
        """Sends an ATT Find By Type Value Request
        """
        self.send_data(ATT_Find_By_Type_Value_Request(
            start=start,
            end=end,
            uuid=type_uuid,
            data=value
        ))

    def  find_by_type_value_response(self, handles: List[ATT_Handle]):
        """Sends an ATT Find By Type Value Response
        """
        self.send_data(ATT_Find_By_Type_Value_Response(
            handles=handles
        ))

    def read_by_type_request(self, start, end, uuid):
        """Sends an ATT Read By Type Request

        :param int start: First requested handle number
        :param int end: Last requested handle number
        :param uuid: 16-bit or 128-bit attribute UUID
        """
        self.send_data(ATT_Read_By_Type_Request(
            start=start,
            end=end,
            uuid=uuid
        ))


    def read_by_type_request_128bit(self, start, end, uuid1, uuid2):
        """Sends an ATT Read By Type Request with 128-bit UUID

        :param int start: First requested handle number
        :param int end: Last requested handle number
        :param uuid1: UUID part 1
        :param uuid2: UUID part 2
        """
        self.send_data(ATT_Read_By_Type_Request_128bit(
            start=start,
            end=end,
            uuid1=uuid1,
            uuid2=uuid2
        ))

    def read_by_type_response(self, item_length, handles):
        """Sends an ATT Read By Type Response

        :param int item_length: Length of a handle item
        :param list handles: List of handles (each item stored on `item_length` bytes)
        """
        self.send_data(ATT_Read_By_Type_Response(
            len=item_length,
            handles=handles
        ))


    def read_request(self, gatt_handle):
        """Sends an ATT Read Request
        """
        self.send_data(ATT_Read_Request(
            gatt_handle=gatt_handle
        ))

    def read_response(self, value):
        """Sends an ATT Read Response
        """
        self.send_data(ATT_Read_Response(
            value=value
        ))


    def read_blob_request(self, handle, offset):
        """Sends an ATT Read Blob Request

        :param int handle: Handle of attribute to read from
        :param int offset: Offset of the first octet to be read
        """
        self.send_data(ATT_Read_Blob_Request(
            gatt_handle=handle,
            offset=offset
        ))

    def read_blob_response(self, value):
        """Sends an ATT Read Blob Response

        :param value: Value read
        """
        self.send_data(ATT_Read_Blob_Response(
            value=value
        ))

    def read_multiple_request(self, handles):
        """Sends an ATT Read Multiple Request

        :param handles: list of handles
        """
        self.send_data(ATT_Read_Multiple_Request(
            handles=handles
        ))

    def read_multiple_response(self, values):
        """Sends an ATT Read Multiple Response

        :param values: List of multiple values
        """
        self.send_data(ATT_Read_Multiple_Response(
            values=values
        ))


    def read_by_group_type_request(self, start, end, uuid):
        """Sends an ATT Read By Group Type Request

        :param int start: First requested handle number
        :param int end: Last requested handle number
        :param uuid: 16-bit or 128-bit group UUID
        """
        self.send_data(ATT_Read_By_Group_Type_Request(
            start=start,
            end=end,
            uuid=uuid
        ))

    def read_by_group_type_response(self, length, data):
        """Sends an ATT Read By Group Type Response

        :param int length: Size of each attribute data
        :param data: List of attribute data
        """
        self.send_data(ATT_Read_By_Group_Type_Response(
            length=length,
            data=data
        ))

    def write_request(self, handle, data):
        """Sends an ATT Write Request

        :param int handle: Attribute handle to write into
        :param data: Data to write
        """
        self.send_data(ATT_Write_Request(
            gatt_handle=handle,
            data=data
        ))

    def write_response(self):
        """Sends an ATT Write Response
        """
        self.send_data(ATT_Write_Response())

    def write_command(self, handle, data):
        """Sends an ATT Write Command
        """
        self.send_data(ATT_Write_Command(
            gatt_handle=handle,
            data=data
        ))


    def prepare_write_request(self, handle, offset, data):
        """Sends an ATT Write Request

        :param int handle: Attribute handle
        :param int offset: Offset of the data to write
        :param data: Data to write
        """
        self.send_data(ATT_Prepare_Write_Request(
            gatt_handle=handle,
            offset=offset,
            data=data
        ))

    def prepare_write_response(self, handle, offset, data):
        """Sends an ATT Write Response

        :param int handle: Attribute handle
        :param int offset: Offset of the data to write
        :param data: Data to write
        """
        self.send_data(ATT_Prepare_Write_Response(
            gatt_handle=handle,
            offset=offset,
            data=data
        ))

    def execute_write_request(self, flags):
        """Sends an ATT Execute Write Request

        :param flags: Flags
        """
        self.send_data(ATT_Execute_Write_Request(
            flags=flags
        ))

    def execute_write_response(self):
        """Sends an ATT Execute Write Response
        """
        self.send_data(ATT_Execute_Write_Response())


    def handle_value_notification(self, handle, value):
        """Sends an ATT Handle Value Notification

        :param int handle: Attribute handle
        :param value: Attribute value
        """
        self.send_data(ATT_Handle_Value_Notification(
            gatt_handle=handle,
            value=value
        ))

    def handle_value_indication(self, handle, value):
        """Sends an ATT Handle Value Indication

        :param int handle: Attribute handle
        :param value: Attribute value
        """
        self.send_data(ATT_Handle_Value_Indication(
            gatt_handle=handle,
            value=value
        ))

    def handle_value_confirmation(self):
        """Sends an ATT Handle Value Confirmation

        Not supported yet
        """
        self.send_data(ATT_Handle_Value_Confirmation())



"""WHAD server tool

This utility implements a server module, allowing to create a TCP proxy
which can be used to access a device remotely.
"""
import logging
from prompt_toolkit import print_formatted_text, HTML

from whad.cli.app import CommandLineApp, ApplicationError
from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady
from whad.device.tcp import TCPSocketConnector
logger = logging.getLogger(__name__)
#logging.basicConfig(level=logging.DEBUG)

class WhadServerApp(CommandLineApp):

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD server tool',
            interface=True,
            commands=False
        )

        self.add_argument(
            '--address',
            '-a',
            dest='address',
            action="store",
            default="127.0.0.1",
            help='IP address to use'
        )

        self.add_argument(
            '--port',
            '-p',
            dest='port',
            action="store",
            default="12345",
            help='Port to use'
        )

    def run(self):
        try:
            # Launch pre-run tasks
            self.pre_run()

            if self.is_piped_interface():
                interface = self.input_interface
            else:
                interface = self.interface

            # Format address and port
            self.address = self.args.address
            self.port = int(self.args.port)

            # We need to have an interface specified
            if interface is not None:
                self.serve(interface)
            else:
                self.error('You have to provide an interface to proxify.')

        except KeyboardInterrupt as keybd:
            self.warning('Server stopped (CTRL-C)')

        if self.server is not None:
            self.server.shutdown()

        # Launch post-run tasks
        self.post_run()

    def serve(self, device):
        '''
        Create a TCP proxy device according to provided address and port and serve forever.
        '''
        print_formatted_text(
            HTML(
                "<ansicyan>[i] Device proxy running on %s:%s </ansicyan>" %
                (self.address, str(self.port))
            )
        )
        self.server = TCPSocketConnector(device, self.address, self.port)
        self.server.serve()

def wserver_main():
    try:
        app = WhadServerApp()
        app.run()
    except ApplicationError as err:
        err.show()
